import numpy as np
from copy import copy

from pint import Quantity
from scipy.signal import fftconvolve

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
from efootprint.constants.units import u


def compute_nb_avg_hourly_occurrences(hourly_occurrences_starts, event_duration):
    """
    Compute the average number of hourly occurrences running in parallel
    for events of a given duration.

    This function replaces a slow Python loop with a single convolution:
    - The loop version repeatedly shifts and sums the hourly occurrences array.
    - Mathematically, this is equivalent to a convolution with a "boxcar"
      (a vector of ones of length = event duration in hours).
    - Using convolution makes the code both much faster and easier to maintain.

    Parameters
    ----------
    hourly_occurrences_starts : ExplainableHourlyQuantities
        The base hourly occurrence pattern (numpy array under `.value`).
    event_duration : ExplainableQuantity
        The duration of one event, converted internally to hours.

    Returns
    -------
    ExplainableHourlyQuantities
        The summed pattern of overlapping events.
    """
    if isinstance(hourly_occurrences_starts, EmptyExplainableObject) or event_duration.magnitude == 0:
        return EmptyExplainableObject(left_parent=hourly_occurrences_starts, right_parent=event_duration)

    # Convert duration to number of hours
    event_duration_in_hours = copy(event_duration.value).to(u.hour).magnitude
    nb_full_hours = int(np.floor(event_duration_in_hours))

    values = hourly_occurrences_starts.value.astype(np.float32, copy=False).magnitude

    # --- Full-hour contribution ---
    if nb_full_hours > 0:
        kernel = np.ones(nb_full_hours, dtype=np.float32)
        result = fftconvolve(values, kernel, mode="full")
    else:
        result = None

    nonfull_duration_rest = event_duration_in_hours - nb_full_hours
    if nonfull_duration_rest > 0:
        if result is None:
            result = (
                    hourly_occurrences_starts.value.astype(np.float32, copy=False) * nonfull_duration_rest)
        else:
            initial_values_padded = np.pad(
                result, (0, 1), constant_values=np.float32(0))
            shifted_values = np.pad(
                hourly_occurrences_starts.value.astype(np.float32, copy=False), (nb_full_hours, 0),
                constant_values=np.float32(0))
            result = initial_values_padded + shifted_values * nonfull_duration_rest

    result = np.maximum(result, 0)  # Clip FFT numerical noise and avoid negative values that could lead to NegativeCumulativeStorageNeedError.

    return ExplainableHourlyQuantities(
        Quantity(result, u.concurrent),
        start_date=hourly_occurrences_starts.start_date,
        left_parent=hourly_occurrences_starts,
        right_parent=event_duration,
        operator="hourly occurrences average"
    )



import math
from abc import abstractmethod
from copy import copy
from typing import List, TYPE_CHECKING

from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.gpu_server import GPUServer
from efootprint.core.hardware.server import Server
from efootprint.core.hardware.server_base import ServerBase
from efootprint.core.usage.compute_nb_occurrences_in_parallel import compute_nb_avg_hourly_occurrences

if TYPE_CHECKING:
    from efootprint.core.usage.usage_pattern import UsagePattern
    from efootprint.core.usage.usage_journey import UsageJourney
    from efootprint.core.usage.usage_journey_step import UsageJourneyStep
    from efootprint.core.hardware.network import Network
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.usage.edge.recurrent_server_need import RecurrentServerNeed


class JobBase(ModelingObject):
    # Mark the class as abstract but not its children when they define a default_values class attribute
    @classmethod
    @abstractmethod
    def default_values(cls):
        pass

    @classmethod
    def attributes_that_can_have_negative_values(cls):
        return ["data_stored"]

    def __init__(self, name: str, data_transferred: ExplainableQuantity, data_stored: ExplainableQuantity,
                 request_duration: ExplainableQuantity, compute_needed: ExplainableQuantity,
                 ram_needed: ExplainableQuantity):
        super().__init__(name)
        self.hourly_occurrences_per_usage_pattern = ExplainableObjectDict()
        self.hourly_avg_occurrences_per_usage_pattern = ExplainableObjectDict()
        self.hourly_data_transferred_per_usage_pattern = ExplainableObjectDict()
        self.hourly_data_stored_per_usage_pattern = ExplainableObjectDict()
        self.hourly_avg_occurrences_across_usage_patterns = EmptyExplainableObject()
        self.hourly_data_transferred_across_usage_patterns = EmptyExplainableObject()
        self.hourly_data_stored_across_usage_patterns = EmptyExplainableObject()
        self.data_transferred = data_transferred.set_label(
            f"Sum of all data uploads and downloads for request {self.name}")
        self.data_stored = data_stored.set_label(f"Data stored by request {self.name}")
        self.request_duration = request_duration.set_label(f"Request duration of {self.name}")
        self.ram_needed = ram_needed.set_label(f"RAM needed to process {self.name}").to(u.MB_ram)
        self.compute_needed = compute_needed.set_label(f"CPU needed to process {self.name}")

    @property
    def calculated_attributes(self) -> List[str]:
        return ["hourly_occurrences_per_usage_pattern", "hourly_avg_occurrences_per_usage_pattern",
                "hourly_data_transferred_per_usage_pattern", "hourly_data_stored_per_usage_pattern",
                "hourly_avg_occurrences_across_usage_patterns", "hourly_data_transferred_across_usage_patterns",
                "hourly_data_stored_across_usage_patterns"]

    @property
    def duration_in_full_hours(self):
        # Use copy not to convert self.request_duration in place
        return ExplainableQuantity(
                math.ceil(copy(self.request_duration.value).to(u.hour).magnitude) * u.dimensionless,
                f"{self.name} duration in full hours")

    # Job objects can be refereenced by UsageJourneySteps or by RecurrentServerNeeds
    @property
    def usage_journey_steps(self) -> List["UsageJourneyStep"]:
        from efootprint.core.usage.usage_journey_step import UsageJourneyStep
        return [obj for obj in self.modeling_obj_containers if isinstance(obj, UsageJourneyStep)]

    @property
    def recurrent_server_needs(self) -> List["RecurrentServerNeed"]:
        from efootprint.core.usage.edge.recurrent_server_need import RecurrentServerNeed
        return [obj for obj in self.modeling_obj_containers if isinstance(obj, RecurrentServerNeed)]

    @property
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        return list(set(sum([rsn.edge_usage_patterns for rsn in self.recurrent_server_needs], start=[])))

    @property
    def usage_journeys(self) -> List["UsageJourney"]:
        return list(set(sum([uj_step.usage_journeys for uj_step in self.usage_journey_steps], start=[])))

    @property
    def web_usage_patterns(self) -> List["UsagePattern"]:
        return list(set(sum([uj_step.usage_patterns for uj_step in self.usage_journey_steps], start=[])))

    @property
    def usage_patterns(self) -> List["UsagePattern| EdgeUsagePattern"]:
        return self.web_usage_patterns + self.edge_usage_patterns

    @property
    def networks(self) -> List["Network"]:
        return list(set(up.network for up in self.usage_patterns))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return self.networks

    def update_dict_element_in_hourly_occurrences_per_usage_pattern(
            self, usage_pattern: "UsagePattern | EdgeUsagePattern"):
        from efootprint.core.usage.usage_pattern import UsagePattern
        if isinstance(usage_pattern, UsagePattern):
            job_occurrences = EmptyExplainableObject()
            delay_between_uj_start_and_job_evt = EmptyExplainableObject()
            for uj_step in usage_pattern.usage_journey.uj_steps:
                for uj_step_job in uj_step.jobs:
                    if uj_step_job == self:
                        job_occurrences += usage_pattern.utc_hourly_usage_journey_starts.return_shifted_hourly_quantities(
                            delay_between_uj_start_and_job_evt)

                delay_between_uj_start_and_job_evt += uj_step.user_time_spent
        else:  # usage_pattern is an EdgeUsagePattern
            job_occurrences = EmptyExplainableObject()
            for recurrent_server_need in self.recurrent_server_needs:
                for recurrent_server_need_job in recurrent_server_need.jobs:
                    if recurrent_server_need_job == self:
                        unitary_hourly_occurrences = recurrent_server_need.unitary_hourly_volume_per_usage_pattern[
                            usage_pattern]
                        job_occurrences += unitary_hourly_occurrences * usage_pattern.nb_edge_usage_journeys_in_parallel

        self.hourly_occurrences_per_usage_pattern[usage_pattern] = job_occurrences.to(u.occurrence).set_label(
            f"Hourly {self.name} occurrences in {usage_pattern.class_as_simple_str} {usage_pattern.name}")

    def update_hourly_occurrences_per_usage_pattern(self):
        self.hourly_occurrences_per_usage_pattern = ExplainableObjectDict()
        for up in self.usage_patterns:
            self.update_dict_element_in_hourly_occurrences_per_usage_pattern(up)

    def update_dict_element_in_hourly_avg_occurrences_per_usage_pattern(
            self, usage_pattern: "UsagePattern | EdgeUsagePattern"):
        hourly_avg_job_occurrences = compute_nb_avg_hourly_occurrences(
            self.hourly_occurrences_per_usage_pattern[usage_pattern], self.request_duration)

        self.hourly_avg_occurrences_per_usage_pattern[usage_pattern] = hourly_avg_job_occurrences.to(u.concurrent).set_label(
            f"Average hourly {self.name} occurrences in {usage_pattern.name}")

    def update_hourly_avg_occurrences_per_usage_pattern(self):
        self.hourly_avg_occurrences_per_usage_pattern = ExplainableObjectDict()
        for up in self.usage_patterns:
            self.update_dict_element_in_hourly_avg_occurrences_per_usage_pattern(up)

    def compute_hourly_data_exchange_for_usage_pattern(
            self, usage_pattern: "UsagePattern | EdgeUsagePattern", data_exchange_type: str):
        data_exchange_type_no_underscore = data_exchange_type.replace("_", " ")

        data_exchange_per_hour = (
                getattr(self, data_exchange_type) * ExplainableQuantity(1 * u.hour, "one hour")
                / self.request_duration
        ).set_label(f"{data_exchange_type_no_underscore} per hour for job {self.name} in {usage_pattern.name}")

        hourly_data_exchange = self.hourly_avg_occurrences_per_usage_pattern[usage_pattern] * data_exchange_per_hour

        return hourly_data_exchange.set_label(
                f"Hourly {data_exchange_type_no_underscore} for {self.name} in {usage_pattern.name}")

    def update_dict_element_in_hourly_data_transferred_per_usage_pattern(
            self, usage_pattern: "UsagePattern | EdgeUsagePattern"):
        self.hourly_data_transferred_per_usage_pattern[usage_pattern] = \
            self.compute_hourly_data_exchange_for_usage_pattern(usage_pattern, "data_transferred")

    def update_hourly_data_transferred_per_usage_pattern(self):
        self.hourly_data_transferred_per_usage_pattern = ExplainableObjectDict()
        for up in self.usage_patterns:
            self.update_dict_element_in_hourly_data_transferred_per_usage_pattern(up)

    def update_dict_element_in_hourly_data_stored_per_usage_pattern(
            self, usage_pattern: "UsagePattern | EdgeUsagePattern"):
        self.hourly_data_stored_per_usage_pattern[usage_pattern] = \
            self.compute_hourly_data_exchange_for_usage_pattern(usage_pattern, "data_stored")

    def update_hourly_data_stored_per_usage_pattern(self):
        self.hourly_data_stored_per_usage_pattern = ExplainableObjectDict()
        for up in self.usage_patterns:
            self.update_dict_element_in_hourly_data_stored_per_usage_pattern(up)

    def sum_calculated_attribute_across_usage_patterns(
            self, calculated_attribute_name: str, calculated_attribute_label: str):
        hourly_calc_attr_summed_across_ups = EmptyExplainableObject()
        for usage_pattern in self.usage_patterns:
            hourly_calc_attr_summed_across_ups += getattr(self, calculated_attribute_name)[usage_pattern]

        return hourly_calc_attr_summed_across_ups.set_label(
                f"Hourly {self.name} {calculated_attribute_label} across usage patterns")

    def update_hourly_avg_occurrences_across_usage_patterns(self):
        self.hourly_avg_occurrences_across_usage_patterns = self.sum_calculated_attribute_across_usage_patterns(
            "hourly_avg_occurrences_per_usage_pattern", "average occurrences").to(u.concurrent)

    def update_hourly_data_transferred_across_usage_patterns(self):
        self.hourly_data_transferred_across_usage_patterns = self.sum_calculated_attribute_across_usage_patterns(
            "hourly_data_transferred_per_usage_pattern", "data transferred")

    def update_hourly_data_stored_across_usage_patterns(self):
        self.hourly_data_stored_across_usage_patterns = self.sum_calculated_attribute_across_usage_patterns(
            "hourly_data_stored_per_usage_pattern", "data stored")


class DirectServerJob(JobBase):
    # Mark the class as abstract but not its children when they define a default_values class attribute
    @classmethod
    @abstractmethod
    def default_values(cls):
        pass

    def __init__(self, name: str, server: ServerBase, data_transferred: ExplainableQuantity,
                 data_stored: ExplainableQuantity, request_duration: ExplainableQuantity,
                 compute_needed: ExplainableQuantity, ram_needed: ExplainableQuantity):
        super().__init__(name, data_transferred, data_stored, request_duration, compute_needed, ram_needed)
        self.server = server
        self.ram_needed.set_label(f"RAM needed on server {self.server.name} to process {self.name}")
        self.compute_needed.set_label(
            f"{str(compute_needed.value.units).replace('_', ' ')}s needed on server {self.server.name} "
            f"to process {self.name}")

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return [self.server] + super().modeling_objects_whose_attributes_depend_directly_on_me


class Job(DirectServerJob):
    default_values =  {
            "data_transferred": SourceValue(150 * u.kB),
            "data_stored": SourceValue(100 * u.kB),
            "request_duration": SourceValue(1 * u.s),
            "compute_needed": SourceValue(0.1 * u.cpu_core),
            "ram_needed": SourceValue(50 * u.MB_ram)
        }

    # __init__ method is copied to change server type.
    def __init__(self, name: str, server: Server, data_transferred: ExplainableQuantity,
                 data_stored: ExplainableQuantity, request_duration: ExplainableQuantity,
                 compute_needed: ExplainableQuantity, ram_needed: ExplainableQuantity):
        super().__init__(name, server, data_transferred, data_stored, request_duration, compute_needed, ram_needed)


class GPUJob(DirectServerJob):
    default_values =  {
            "data_transferred": SourceValue(150 * u.kB),
            "data_stored": SourceValue(100 * u.kB),
            "request_duration": SourceValue(1 * u.s),
            "compute_needed": SourceValue(1 * u.gpu),
            "ram_needed": SourceValue(50 * u.MB_ram)
        }

    def __init__(self, name: str, server: GPUServer, data_transferred: ExplainableQuantity,
                 data_stored: ExplainableQuantity, request_duration: ExplainableQuantity,
                 compute_needed: ExplainableQuantity, ram_needed: ExplainableQuantity):
        super().__init__(name, server, data_transferred, data_stored, request_duration, compute_needed, ram_needed)



from typing import List, TYPE_CHECKING

from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.usage.edge.edge_function import EdgeFunction

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.hardware.edge.edge_device import EdgeDevice
    from efootprint.core.usage.edge.recurrent_edge_device_need import RecurrentEdgeDeviceNeed
    from efootprint.core.usage.edge.recurrent_server_need import RecurrentServerNeed
    from efootprint.core.usage.job import JobBase


class EdgeUsageJourney(ModelingObject):
    default_values = {
        "usage_span": SourceValue(6 * u.year)
    }

    def __init__(self, name: str, edge_functions: List[EdgeFunction], usage_span: ExplainableQuantity):
        super().__init__(name)
        self.edge_functions = edge_functions
        self.usage_span = usage_span.set_label(f"Usage span of {self.name}")

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List["EdgeUsagePattern"] | List[EdgeFunction]:
        if self.edge_usage_patterns:
            return self.edge_usage_patterns
        return self.edge_functions

    @property
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        return self.modeling_obj_containers

    @property
    def recurrent_edge_device_needs(self) -> List["RecurrentEdgeDeviceNeed"]:
        return list(set(sum([ef.recurrent_edge_device_needs for ef in self.edge_functions], start=[])))

    @property
    def recurrent_server_needs(self) -> List["RecurrentServerNeed"]:
        return list(set(sum([ef.recurrent_server_needs for ef in self.edge_functions], start=[])))

    @property
    def jobs(self) -> List["JobBase"]:
        return list(set(sum([rsn.jobs for rsn in self.recurrent_server_needs], start=[])))

    @property
    def edge_devices(self) -> List["EdgeDevice"]:
        return list(set([edge_need.edge_device for edge_need in self.recurrent_edge_device_needs]))



from efootprint.abstract_modeling_classes.explainable_dict import ExplainableDict
from efootprint.abstract_modeling_classes.explainable_object_base_class import Source, ExplainableObject
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.builders.hardware.boaviztapi_utils import call_boaviztapi
from efootprint.constants.sources import Sources
from efootprint.constants.units import u
from efootprint.core.hardware.server_base import ServerTypes, ServerBase
from efootprint.core.hardware.storage import Storage


class BoaviztaServerFromConfig(ServerBase):
    def _abc_marker(self):
        pass  # silent override

    default_values = {
            "server_type": ServerTypes.on_premise(),
            "nb_of_cpu_units": SourceValue(1 * u.dimensionless),
            "nb_of_cores_per_cpu_unit": SourceValue(1 * u.dimensionless),
            "nb_of_ram_units": SourceValue(1 * u.dimensionless),
            "ram_quantity_per_unit": SourceValue(1 * u.GB),
            "average_carbon_intensity": SourceValue(400 * u.g / u.kWh),
            "lifespan": SourceValue(6 * u.year),
            "idle_power": SourceValue(0 * u.W),
            "power_usage_effectiveness": SourceValue(1.4 * u.dimensionless),
            "utilization_rate": SourceValue(0.7 * u.dimensionless),
            "base_ram_consumption": SourceValue(0 * u.GB_ram),
            "base_compute_consumption": SourceValue(0 * u.cpu_core),
            "storage": Storage.ssd(storage_capacity=SourceValue(32 * u.GB)),
            "fixed_nb_of_instances": EmptyExplainableObject()
        }
    
    def __init__(self, name: str, server_type: ExplainableObject, nb_of_cpu_units: ExplainableQuantity,
                 nb_of_cores_per_cpu_unit: ExplainableQuantity,
                 nb_of_ram_units: ExplainableQuantity, ram_quantity_per_unit: ExplainableQuantity,
                 average_carbon_intensity: ExplainableQuantity, lifespan: ExplainableQuantity,
                 idle_power: ExplainableQuantity, power_usage_effectiveness: ExplainableQuantity,
                 utilization_rate: ExplainableQuantity, base_ram_consumption: ExplainableQuantity,
                 base_compute_consumption: ExplainableQuantity, storage: Storage,
                 fixed_nb_of_instances: ExplainableQuantity | EmptyExplainableObject | None = None):
        super().__init__(
            name, server_type=server_type, carbon_footprint_fabrication=SourceValue(0 * u.kg),
            power=SourceValue(0 * u.kg), lifespan=lifespan, idle_power=idle_power, ram=SourceValue(0 * u.GB),
            compute=SourceValue(0 * u.cpu_core), power_usage_effectiveness=power_usage_effectiveness,
            average_carbon_intensity=average_carbon_intensity,
            utilization_rate=utilization_rate, base_ram_consumption=base_ram_consumption,
            base_compute_consumption=base_compute_consumption, fixed_nb_of_instances=fixed_nb_of_instances,
            storage=storage)
        self.nb_of_cpu_units = nb_of_cpu_units
        self.nb_of_cores_per_cpu_unit = nb_of_cores_per_cpu_unit
        self.nb_of_ram_units = nb_of_ram_units
        self.ram_quantity_per_unit = ram_quantity_per_unit
        self.impact_url = "https://api.boavizta.org/v1/server/"
        # TODO: allow for archetype setting
        self.params = {"verbose": "true", "archetype": "platform_compute_medium", "criteria": ["gwp"]}
        self.impact_source = Source(
            name="Boavizta API servers",
            link=f"{self.impact_url}?{'&'.join([key + '=' + str(self.params[key]) for key in self.params])}")
        self.cpu_config = EmptyExplainableObject()
        self.ram_config = EmptyExplainableObject()
        self.api_call_response = EmptyExplainableObject()
        
    @property
    def calculated_attributes(self):
        return (["cpu_config", "ram_config", "api_call_response", "carbon_footprint_fabrication", "power", "ram", 
                 "compute"] + super().calculated_attributes)

    @property
    def attributes_that_shouldnt_trigger_update_logic(self):
        return super().attributes_that_shouldnt_trigger_update_logic + ["impact_url", "params", "impact_source"]

    # TODO: Introduce a notion of compatible storages

    def after_init(self):
        super().after_init()
        # TODO: uncomment line after fixing the API call structure
        # self.compute_calculated_attributes()
    
    def update_cpu_config(self):
        cpu_config_dict = {"units": self.nb_of_cpu_units.to(u.dimensionless).magnitude,
                           "core_units": self.nb_of_cores_per_cpu_unit.to(u.dimensionless).magnitude}

        self.cpu_config = ExplainableObject(
            cpu_config_dict, label=f"{self.name} cpu config", left_parent=self.nb_of_cpu_units,
            right_parent=self.nb_of_cores_per_cpu_unit, operator="combined in Boavizta API with",
            source=Sources.USER_DATA)

    def update_ram_config(self):
        ram_config_dict = {"units": self.nb_of_ram_units.to(u.dimensionless).magnitude,
                           "capacity": self.ram_quantity_per_unit.to(u.GB).magnitude}

        self.ram_config = ExplainableObject(
            ram_config_dict, label=f"{self.name} ram config", left_parent=self.nb_of_ram_units,
            right_parent=self.ram_quantity_per_unit, operator="combined in Boavizta API with",
            source=Sources.USER_DATA)

    def update_api_call_response(self):
        # TODO: fix api call structure
        api_call_data = {
            "model": {"type": "rack"},
            "configuration": {"cpu": self.cpu_config.value, "ram": self.ram_config.value}
        }
        self.api_call_response = ExplainableDict(
            call_boaviztapi(url=self.impact_url, params=self.params, json=api_call_data, method="POST"),
            label=f"{self.name} api call data", left_parent=self.cpu_config, right_parent=self.ram_config,
            operator="combined in Boavizta API data with", source=Sources.USER_DATA)

    def update_carbon_footprint_fabrication(self):
        total_fabrication_footprint_storage_included = ExplainableQuantity(
            self.api_call_response.value["impacts"]["gwp"]["embedded"]["value"] * u.kg,
            f"Total {self.name} fabrication footprint storage included", 
            left_parent=self.api_call_response, operator="data extraction from",
            source=self.impact_source)

        storage_spec = self.api_call_response.value["verbose"].get("SSD-1", self.api_call_response.value["verbose"].get("HDD-1", None))
        if storage_spec is None:
            raise ValueError("Both SSD and HDD storage found in the server impact data. This is not implemented yet")
        
        full_storage_carbon_footprint_fabrication = ExplainableQuantity(
            storage_spec["impacts"]["gwp"]["embedded"]["value"] * u.kg, f"Total {self.name} fabrication footprint",
            left_parent=self.api_call_response, operator="data extraction from",
            source=self.impact_source)
        
        total_fabrication_footprint_storage_excluded = (
                total_fabrication_footprint_storage_included - full_storage_carbon_footprint_fabrication)
        
        self.carbon_footprint_fabrication = total_fabrication_footprint_storage_excluded.set_label(
            f"Fabrication footprint of {self.name}")
        
    def update_power(self):
        average_power_value = self.api_call_response.value["verbose"]["avg_power"]["value"]
        average_power_unit = self.api_call_response.value["verbose"]["avg_power"]["unit"]
        use_time_ratio = self.api_call_response.value["verbose"]["use_time_ratio"]["value"]

        assert average_power_unit == "W", f"Unexpected power unit {average_power_unit}"
        assert float(use_time_ratio) == 1, f"Unexpected use time ratio {use_time_ratio}"
        
        self.power = ExplainableQuantity(
            average_power_value * u.W, f"{self.name} power", left_parent=self.api_call_response,
            operator="data extraction from", source=self.impact_source)
        
    def update_ram(self):
        ram_spec = self.api_call_response.value["verbose"]["RAM-1"]

        self.ram = ExplainableQuantity(
            ram_spec["units"]["value"] * ram_spec["capacity"]["value"] * u.GB_ram, f"{self.name} ram",
            left_parent=self.api_call_response, operator="data extraction from", source=self.impact_source)
        
    def update_compute(self):
        cpu_spec = self.api_call_response.value["verbose"]["CPU-1"]

        self.compute = ExplainableQuantity(
            cpu_spec["units"]["value"] * cpu_spec["core_units"]["value"] * u.cpu_core, f"{self.name} compute",
            left_parent=self.api_call_response, operator="data extraction from", source=self.impact_source)
        

class BoaviztaStorageFromConfig(Storage):
    default_values =  {
            "idle_power": SourceValue(0 * u.W),
            "data_replication_factor": SourceValue(1 * u.dimensionless),
            "data_storage_duration": SourceValue(1 * u.year),
            "base_storage_need": SourceValue(0 * u.GB)
        }
    # TODO: create StorageBase class and introduce compatible_storages in servers

    def __init__(self, name: str, idle_power: ExplainableQuantity, data_replication_factor: ExplainableQuantity,
                 data_storage_duration: ExplainableQuantity, base_storage_need: ExplainableQuantity):
        super().__init__(
            name, carbon_footprint_fabrication_per_storage_capacity=SourceValue(0 * u.kg / u.TB),
            power_per_storage_capacity=SourceValue(0 * u.W / u.TB),
            lifespan=SourceValue(0 * u.year), idle_power=idle_power, storage_capacity=SourceValue(0 * u.TB), 
            data_replication_factor=data_replication_factor, data_storage_duration=data_storage_duration, 
            base_storage_need=base_storage_need)
        
        self.storage_type = EmptyExplainableObject()

    @property
    def calculated_attributes(self):
        return (["storage_type", "storage_capacity", "fixed_nb_of_instances",
                 "carbon_footprint_fabrication_per_storage_capacity",
                 "power_per_storage_capacity", "lifespan"] + super().calculated_attributes)

    def update_storage_type(self):
        storage_type = None
        if ("SSD-1" in self.server.api_call_response.value["verbose"]
                and "HDD-1" in self.server.api_call_response.value["verbose"]):
            raise ValueError("Both SSD and HDD storage found in the server impact data. This is not implemented yet")
        elif "SSD-1" in self.server.api_call_response.value["verbose"]:
            storage_type = "SSD"
        elif "HDD-1" in self.server.api_call_response.value["verbose"]:
            storage_type = "HDD"

        self.storage_type = ExplainableObject(
            storage_type, label=f"{self.name} storage type", left_parent=self.server.api_call_response.value,
            operator="data extraction from", source=self.server.impact_source)

    def update_storage_capacity(self):
        storage_spec = self.server.api_call_response.value["verbose"][f"{self.storage_type.value}-1"]
        storage_unit = getattr(u, storage_spec["capacity"]["unit"])
        
        self.storage_capacity = ExplainableQuantity(
            storage_spec["capacity"]["value"] * storage_unit, f"{self.name} storage capacity", 
            left_parent=self.server.api_call_response, operator="data extraction", source=self.server.impact_source)

    def update_fixed_nb_of_instances(self):
        storage_spec = self.server.api_call_response.value["verbose"][f"{self.storage_type.value}-1"]
        nb_units = ExplainableQuantity(
            storage_spec["units"]["value"] * u.dimensionless, label=f"Number of {self.name} storage instances",
            left_parent=self.server.api_call_response, operator="data extraction",
            source=self.server.impact_source)

        self.fixed_nb_of_instances = nb_units.set_label(f"Fixed number of {self.name} storage instances")

    def update_carbon_footprint_fabrication_per_storage_capacity(self):
        storage_spec = self.server.api_call_response.value["verbose"][f"{self.storage_type.value}-1"]
        full_storage_carbon_footprint_fabrication = ExplainableQuantity(
            storage_spec["impacts"]["gwp"]["embedded"]["value"] * u.kg, left_parent=self.storage_type,
            source=self.server.impact_source,
            label=f"Total {self.name} fabrication footprint")
        
        self.carbon_footprint_fabrication_per_storage_capacity = (
                full_storage_carbon_footprint_fabrication / (self.fixed_nb_of_instances * self.storage_capacity)
        ).set_label(f"Fabrication footprint of one {self.name} storage instance")

    def update_power_per_storage_capacity(self):
        power_per_storage_capacity = None
        if self.storage_type.value == "SSD":
            power_per_storage_capacity = SourceValue(1.3 * u.W / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY)
        elif self.storage_type.value == "HDD":
            power_per_storage_capacity = SourceValue(4.2 * u.W / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY)

        self.power_per_storage_capacity = (
            power_per_storage_capacity.generate_explainable_object_with_logical_dependency(
                self.storage_type).set_label(f"{self.name} power per storage capacity"))

    def update_lifespan(self):
        lifespan = None
        if self.storage_type.value == "SSD":
            lifespan = SourceValue(6 * u.year)
        elif self.storage_type.value == "HDD":
            lifespan = SourceValue(4 * u.year)

        self.lifespan = lifespan.generate_explainable_object_with_logical_dependency(self.storage_type).set_label(
            f"{self.name} lifespan")



import re
from typing import List

from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue, Sources, SourceObject
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.builders.services.service_base_class import Service
from efootprint.builders.services.service_job_base_class import ServiceJob
from efootprint.constants.units import u
from efootprint.core.hardware.server import Server


class VideoStreaming(Service):
    default_values =  {
            "base_ram_consumption": SourceValue(2 * u.GB_ram, source=Sources.HYPOTHESIS),
            "bits_per_pixel": SourceValue(0.1 * u.dimensionless, source=Sources.HYPOTHESIS),
            "static_delivery_cpu_cost": SourceValue(4 * u.cpu_core / (u.GB / u.s), source=Sources.HYPOTHESIS),
            "ram_buffer_per_user": SourceValue(50 * u.MB_ram, source=Sources.HYPOTHESIS),
            }

    def __init__(self, name: str, server: Server, base_ram_consumption: ExplainableQuantity,
                 bits_per_pixel: ExplainableQuantity, static_delivery_cpu_cost: ExplainableQuantity,
                 ram_buffer_per_user: ExplainableQuantity):
        super().__init__(name, server)
        self.base_ram_consumption = base_ram_consumption.set_label(
            f"{self.name} OS and streaming software base RAM consumption")
        self.bits_per_pixel = bits_per_pixel.set_label(f"{self.name} bits per pixel")
        self.static_delivery_cpu_cost = static_delivery_cpu_cost.set_label(f"{self.name} CPU cost per static stream")
        self.ram_buffer_per_user = ram_buffer_per_user.set_label(f"{self.name} RAM buffer size per user")


class VideoStreamingJob(ServiceJob):
    default_values =  {
            "resolution": SourceObject("1080p (1920 x 1080)"),
            "video_duration": SourceValue(1 * u.hour),
            "refresh_rate": SourceValue(30 * u.dimensionless / u.s),
            "data_stored": SourceValue(0 * u.MB),
        }

    list_values =  {"resolution": [
            SourceObject("480p (640 x 480)"), SourceObject("720p (1280 x 720)"), SourceObject("1080p (1920 x 1080)"),
            SourceObject("1440p (2560 x 1440)"), SourceObject("2K (2048 x 1080)"),
            SourceObject("4K (3840 x 2160)"), SourceObject("8K (7680 x 4320)")]
        }

    def __init__(self, name: str, service: VideoStreaming, resolution: ExplainableObject,
                 video_duration: ExplainableQuantity, refresh_rate: ExplainableQuantity,
                 data_stored: ExplainableQuantity):
        super().__init__(name or f"{resolution} streaming on {service.name}",
                         service, SourceValue(0 * u.kB), data_stored,
                         SourceValue(0 * u.s), SourceValue(0 * u.cpu_core), SourceValue(0 * u.GB))
        self.video_duration = video_duration.set_label(f"{self.name} video duration")
        self.resolution = resolution.set_label(f"{self.name} resolution")
        self.refresh_rate = refresh_rate.set_label(f"{self.name} frames per second")
        self.dynamic_bitrate = EmptyExplainableObject()

    @property
    def calculated_attributes(self) -> List[str]:
        return (["request_duration", "dynamic_bitrate", "data_transferred", "compute_needed", "ram_needed"]
                + super().calculated_attributes)

    def update_request_duration(self):
        self.request_duration = self.video_duration.copy().set_label(f"{self.name} request duration")

    def update_dynamic_bitrate(self):
        match = re.search(r"\((\d+)\s*x\s*(\d+)\)", self.resolution.value)
        if not match:
            raise ValueError(f"Invalid resolution format: {self.resolution.value}")
        width, height = map(int, match.groups())
        pixel_count = ExplainableQuantity(
            width * height * u.dimensionless, f"pixel count for resolution {self.resolution}",
            left_parent=self.resolution, operator="pixel count computation", source=Sources.USER_DATA)

        self.dynamic_bitrate = (pixel_count * self.service.bits_per_pixel * self.refresh_rate
                                ).to(u.MB / u.s).set_label(f"{self.name} dynamic bitrate")

    def update_data_transferred(self):
        self.data_transferred = (self.request_duration * self.dynamic_bitrate).to(u.GB).set_label(
            f"{self.name} data transferred")

    def update_compute_needed(self):
        self.compute_needed = (self.service.static_delivery_cpu_cost * self.dynamic_bitrate).to(u.cpu_core).set_label(
            f"{self.name} CPU needed")

    def update_ram_needed(self):
        self.ram_needed = self.service.ram_buffer_per_user.copy().set_label(f"{self.name} RAM needed")



from typing import TYPE_CHECKING

from pint import Unit, Quantity
import numpy as np

from efootprint.abstract_modeling_classes.explainable_object_base_class import (
    ExplainableObject, Source)
from efootprint.constants.units import u, get_unit
from efootprint.logger import logger
from efootprint.abstract_modeling_classes.aggregation_utils import validate_timeseries_unit

if TYPE_CHECKING:
    from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
    from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
    from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
    from efootprint.abstract_modeling_classes.explainable_timezone import ExplainableTimezone


@ExplainableObject.register_subclass(lambda d: "recurring_values" in d and "unit" in d)
class ExplainableRecurrentQuantities(ExplainableObject):
    __slots__ = (
        '_ExplainableQuantity',
        '_EmptyExplainableObject',
    )

    @classmethod
    def from_json_dict(cls, d):
        source = Source.from_json_dict(d.get("source")) if d.get("source") else None
        value = Quantity(np.array(eval(d["recurring_values"]), dtype=np.float32), get_unit(d["unit"]))

        return cls(value, label=d["label"], source=source)

    def __init__(
            self, value: Quantity, label: str = None,
            left_parent: ExplainableObject = None, right_parent: ExplainableObject = None, operator: str = None,
            source: Source = None):
        from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
        from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
        self._ExplainableQuantity = ExplainableQuantity
        self._EmptyExplainableObject = EmptyExplainableObject
        if isinstance(value, Quantity):
            validate_timeseries_unit(value, label)
            if value.magnitude.dtype != np.float32:
                logger.info(
                    f"converting value {label} to float32. This is surprising, a casting to np.float32 is probably "
                    f"missing somewhere.")
                value = value.magnitude.astype(np.float32, copy=False) * value.units
            super().__init__(value, label, left_parent, right_parent, operator, source)
        else:
            raise ValueError(
                f"ExplainableRecurrentQuantities values must be Pint Quantities of numpy arrays, got {type(value)}"
            )

    def __eq__(self, other):
        if isinstance(other, ExplainableRecurrentQuantities):
            return np.allclose(self.value, other.value, rtol=1e-06, atol=1e-06)
        return False

    def __add__(self, other):
        if isinstance(other, (int, float)) and other == 0:
            return ExplainableRecurrentQuantities(self.value, label=self.label, left_parent=self, operator="")
        elif isinstance(other, self._EmptyExplainableObject):
            return ExplainableRecurrentQuantities(
                self.value, label=self.label, left_parent=self, right_parent=other, operator="+")
        elif isinstance(other, ExplainableRecurrentQuantities):
            if len(self.value) != len(other.value):
                raise ValueError(f"Cannot add ExplainableRecurrentQuantities with different lengths: "
                                 f"{len(self.value)} vs {len(other.value)}")
            return ExplainableRecurrentQuantities(
                self.value + other.value, label=None, left_parent=self, right_parent=other, operator="+")
        elif isinstance(other, self._ExplainableQuantity):
            return ExplainableRecurrentQuantities(
                self.value + other.value, label=None, left_parent=self, right_parent=other, operator="+")
        else:
            raise ValueError(f"Can only add another ExplainableRecurrentQuantities, ExplainableQuantity, scalar 0, "
                             f"or EmptyExplainableObject, not {type(other)}")

    def to(self, unit_to_convert_to: Unit):
        self.value = self.value.to(unit_to_convert_to)
        validate_timeseries_unit(self.value, self.label)

        return self

    def generate_explainable_object_with_logical_dependency(self, explainable_condition: "ExplainableObject"):
        return self.__class__(
            value=self.value, label=self.label, left_parent=self,
            right_parent=explainable_condition, operator="logically dependent on")

    def __round__(self, round_level):
        return ExplainableRecurrentQuantities(
            np.round(self.value, round_level).astype(np.float32, copy=False), label=self.label,
            left_parent=self, operator=f"rounded to {round_level} decimals", source=self.source
        )

    def round(self, round_level):
        self.value = np.round(self.value, round_level).astype(np.float32, copy=False)

        return self

    @property
    def unit(self):
        return self.value.units

    @property
    def magnitude(self):
        return self.value.magnitude

    @property
    def value_as_float_list(self):
        return self.magnitude.tolist()

    @property
    def plot_aggregation_strategy(self) -> str:
        """Determine how to aggregate recurring values into daily values for plotting."""
        from efootprint.abstract_modeling_classes.aggregation_utils import get_plot_aggregation_strategy
        return get_plot_aggregation_strategy(self.unit)

    def copy(self):
        return ExplainableRecurrentQuantities(
            self.value.copy(), label=self.label, left_parent=self, operator="duplicate")

    def generate_hourly_quantities_over_timespan(
            self, timespan_hourly_quantities: "ExplainableHourlyQuantities", 
            local_timezone: "ExplainableTimezone"):
        """
        Generate an ExplainableHourlyQuantities over the timespan of the input ExplainableHourlyQuantities.
        
        The recurring values represent a canonical local timezone week (7 * 24 = 168 values).
        The timespan_hourly_quantities start_date is expressed in the local timezone but its start_date is not timezone aware.
        The method aligns the canonical week pattern with the local timespan and outputs in UTC.
        
        Args:
            timespan_hourly_quantities: ExplainableHourlyQuantities that defines the timespan and start date (local timezone)
            local_timezone: ExplainableTimezone that defines the local timezone
            
        Returns:
            ExplainableHourlyQuantities with values generated from the recurring pattern (UTC)
        """
        from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities

        # Validate canonical week length
        if len(self.value) != 168:
            raise ValueError(
                f"ExplainableRecurrentQuantities must have exactly 168 values (7*24 hours), got {len(self.value)}"
            )

        if timespan_hourly_quantities.start_date.tzinfo is None:
            start_date_local = local_timezone.value.localize(timespan_hourly_quantities.start_date)
        else:
            start_date_local = timespan_hourly_quantities.start_date.astimezone(local_timezone.value)
        timespan_length = len(timespan_hourly_quantities.value)
        start_offset_in_local_week = start_date_local.weekday() * 24 + start_date_local.hour

        # Vectorized index mapping into canonical local week (168 hours)
        indices = (start_offset_in_local_week + np.arange(timespan_length)) % 168

        # Vectorized extraction of values
        week_values = self.value.magnitude  # is numpy array
        output_values = np.take(week_values, indices)

        result_quantity = Quantity(output_values.astype(np.float32), self.unit)

        local_timezone_expanded = ExplainableHourlyQuantities(
            result_quantity,
            start_date=start_date_local,
            left_parent=self,
            right_parent=timespan_hourly_quantities,
            operator="expanded over timespan",
        )

        return local_timezone_expanded.convert_to_utc(local_timezone).set_label(
            f"{self.label} expanded over {timespan_hourly_quantities.label} timespan (UTC)" if self.label else None,
        )

    def to_json(self, save_calculated_attributes=False):
        output_dict = {
                "recurring_values": str(self.magnitude.tolist()),
                "unit": str(self.unit),
            }

        output_dict.update(super().to_json(save_calculated_attributes))

        return output_dict

    def __repr__(self):
        return str(self)

    def __str__(self):
        def _round_series_values(input_series: np.array):
            return [str(round(hourly_value.magnitude, 2)) for hourly_value in input_series]

        compact_unit = "{:~}".format(self.unit)
        nb_of_values = len(self.value)
        if nb_of_values < 30:
            rounded_values = _round_series_values(self.value)
            str_rounded_values = "[" + ", ".join(rounded_values) + "]"
        else:
            first_vals = _round_series_values(self.value[:10])
            last_vals = _round_series_values(self.value[-10:])
            str_rounded_values = "first 10 vals [" + ", ".join(first_vals) \
                                 + "],\n    last 10 vals [" + ", ".join(last_vals) + "]"

        return f"{nb_of_values} values in {compact_unit}:\n    {str_rounded_values}"



from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.object_linked_to_modeling_obj import ObjectLinkedToModelingObj


class ContextualModelingObjectAttribute(ObjectLinkedToModelingObj):
    __slots__ = ('_value',)

    def __init__(self, value: ModelingObject, modeling_obj_container=None, attr_name_in_mod_obj_container=None):
        super().__init__()
        if isinstance(value, ContextualModelingObjectAttribute):
            self._value = value._value
        elif isinstance(value, ModelingObject):
            self._value = value
        else:
            raise TypeError(f"Expected a ModelingObject or ContextualModelingObjectAttribute, got {type(value)}")
        self._value.add_to_contextual_modeling_obj_containers(self)
        self.modeling_obj_container = modeling_obj_container
        self.attr_name_in_mod_obj_container = attr_name_in_mod_obj_container

    @property
    def id(self):
        return self._value.id

    def __getattr__(self, attr):
        return getattr(self._value, attr)  # Use `getattr` instead of `__getattr__`

    def __getattribute__(self, name):
        if name == "__dict__":
            # Redirect `__dict__` access to the wrapped object's `__dict__`
            return self._value.__dict__
        return super().__getattribute__(name)

    def __setattr__(self, name, input_value, *args, **kwargs):
        if name in ["_value", "modeling_obj_container", "attr_name_in_mod_obj_container",
                    "former_modeling_obj_container_id", "former_attr_name_in_mod_obj_container",
                    "_cached_id", "_cached_full_str_tuple_id", "_cached_attribute_id",
                    "_cached_dict_container", "_cached_key_in_dict", "_cached_list_container",
                    "_cached_indexes_in_list"]:
            # If setting a class attribute, use the superclass's __setattr__
            super().__setattr__(name, input_value)
        else:
            setattr(self._value, name, input_value)  # Use `setattr` instead of `__setattr__`

    def __getitem__(self, key):
        return self._value[key]

    def __repr__(self):
        return repr(self._value)

    def __str__(self):
        return str(self._value)

    def __call__(self, *args, **kwargs):
        return self._value(*args, **kwargs)

    def __eq__(self, other):
        return self._value == other

    def __hash__(self):
        return hash(self._value)  # Use built-in hash directly



import os
import time
from collections import OrderedDict

import orjson
import requests

from efootprint.logger import logger


_BOAVIZTAPI_CACHE_MAX_BYTES = 50 * 1024 * 1024
_BOAVIZTAPI_CACHE_TTL_SECONDS = 7 * 24 * 60 * 60
_boaviztapi_cache = OrderedDict()
_boaviztapi_cache_size_bytes = 0


def _make_cache_key(url, method, params, json_payload):
    cache_payload = {"url": url, "method": method, "params": params or {}, "json": json_payload}
    return orjson.dumps(cache_payload, option=orjson.OPT_SORT_KEYS, default=str)


def _estimate_cache_entry_size(cache_key, value):
    try:
        value_bytes = len(orjson.dumps(value, option=orjson.OPT_SORT_KEYS, default=str))
    except TypeError:
        value_bytes = len(repr(value).encode("utf-8"))
    return len(cache_key) + value_bytes


def _cache_get(cache_key):
    entry = _boaviztapi_cache.get(cache_key)
    if entry is None:
        return None
    value, entry_size, expires_at = entry
    if expires_at <= time.time():
        _boaviztapi_cache.pop(cache_key, None)
        global _boaviztapi_cache_size_bytes
        _boaviztapi_cache_size_bytes -= entry_size
        return None
    return value


def _cache_set(cache_key, value):
    global _boaviztapi_cache_size_bytes
    entry_size = _estimate_cache_entry_size(cache_key, value)
    if entry_size > _BOAVIZTAPI_CACHE_MAX_BYTES:
        return
    now = time.time()
    for key, (_, old_size, expires_at) in list(_boaviztapi_cache.items()):
        if expires_at <= now:
            _boaviztapi_cache.pop(key, None)
            _boaviztapi_cache_size_bytes -= old_size
    while _boaviztapi_cache and (_boaviztapi_cache_size_bytes + entry_size) > _BOAVIZTAPI_CACHE_MAX_BYTES:
        _, (_, old_size, _) = _boaviztapi_cache.popitem(last=False)
        _boaviztapi_cache_size_bytes -= old_size
    _boaviztapi_cache[cache_key] = (value, entry_size, now + _BOAVIZTAPI_CACHE_TTL_SECONDS)
    _boaviztapi_cache_size_bytes += entry_size


def call_boaviztapi(url, method="GET", params=None, json=None):
    if os.getenv("USE_BOAVIZTAPI_PACKAGE"):
        return call_boaviztapi_from_package_dependency(url, method, params, json)
    cache_key = _make_cache_key(url, method, params, json)
    cached_response = _cache_get(cache_key)
    if cached_response is not None:
        logger.info(f"Fetched {method} {url} params {params} Boavizta API response from cache.")
        return cached_response
    try:
        response = call_boaviztapi_from_web_request(url, method, params, json)
        _cache_set(cache_key, response)
        return response
    except Exception as e:
        logger.warning(f"Boavizta API call failed with error {e}. Trying to call Boavizta API via package dependency.")
        response = call_boaviztapi_from_package_dependency(url, method, params, json)
        _cache_set(cache_key, response)
        return response


def call_boaviztapi_from_web_request(url, method="GET", params=None, json=None):
    params = params or {}
    from time import perf_counter
    start = perf_counter()
    headers = {'accept': 'application/json'}
    response = None
    if method == "GET":
        response = requests.get(url, headers=headers, params=params)
    elif method == "POST":
        headers["Content-Type"] = "application/json"
        response = requests.post(url, headers=headers, params=params, json=json)

    if response.status_code == 200:
        logger.info(f"Called {method} {url} with params {params} in {int((perf_counter() - start) * 1000)} ms.")
        return response.json()
    else:
        raise ValueError(
            f"{method} request to {url} with params {params} failed with status code {response.status_code}")


def call_boaviztapi_from_package_dependency(url, method="GET", params=None, json=None):
    params = params or {}
    import asyncio
    import inspect
    import warnings
    from scipy.optimize import OptimizeWarning

    warnings.simplefilter("ignore", ResourceWarning)
    warnings.simplefilter("ignore", OptimizeWarning)
    warnings.simplefilter("ignore", DeprecationWarning)
    warnings.simplefilter("ignore", UserWarning)

    from boaviztapi.routers.cloud_router import instance_cloud_impact, server_get_all_provider_name, \
        server_get_all_archetype_name as server_get_all_archetype_name_cloud
    from boaviztapi.routers.server_router import (
        server_impact_from_model, server_get_all_archetype_name as server_get_all_archetype_name_server,
        get_archetype_config)

    url_method_mapping = {
        "https://api.boavizta.org/v1/cloud/instance/all_providers": server_get_all_provider_name,
        "https://api.boavizta.org/v1/cloud/instance/all_instances": server_get_all_archetype_name_cloud,
        "https://api.boavizta.org/v1/cloud/instance": instance_cloud_impact,
        "https://api.boavizta.org/v1/server/": server_impact_from_model,
        "https://api.boavizta.org/v1/server/archetypes": server_get_all_archetype_name_server,
        "https://api.boavizta.org/v1/server/archetype_config": get_archetype_config,
    }

    if url in url_method_mapping:
        method = url_method_mapping[url]
        if "criteria" in inspect.signature(method).parameters:
            params["criteria"] = params.get("criteria", ["gwp"])
        return asyncio.run(method(**params))
    else:
        raise ValueError(
            f"URL {url} is not in the list of available urls: {list(url_method_mapping.keys())}. Please provide a valid "
            f"URL, update the list of urls in call_boaviztapi_from_package_dependency")


def print_archetypes_and_their_configs():
    for archetype in call_boaviztapi('https://api.boavizta.org/v1/server/archetypes'):
        config = call_boaviztapi(
            url="https://api.boavizta.org/v1/server/archetype_config", params={"archetype": archetype})
        impact = call_boaviztapi(
            url="https://api.boavizta.org/v1/server/", params={"archetype": archetype})

        if "default" in config['CPU']['core_units']:
            nb_cpu_core_units = config['CPU']['core_units']['default']
        elif "core_units" in impact["verbose"]['CPU-1']:
            nb_cpu_core_units = impact["verbose"]['CPU-1']['core_units']['value']
        else:
            nb_cpu_core_units = 1

        nb_ssd_units = config['SSD']["units"].get('default', 0)
        nb_hdd_units = config['HDD']["units"].get('default', 0)

        if nb_hdd_units > 0 and nb_ssd_units > 0:
            raise ValueError(
                f"Archetype {archetype} has both SSD and HDD, please check and delete this exception raising if ok")
        storage_type = "SSD"
        if nb_hdd_units > 0:
            storage_type = "HDD"
        nb_storage_units = config[storage_type]["units"]['default']

        print(
            f"{archetype}: type {config['CASE']['case_type']['default']},\n"
            f"    {config['CPU']['units']['default']} cpu units with {nb_cpu_core_units} core units,\n"
            f"    {config['RAM']['units']['default']} RAM units with {config['RAM']['capacity']['default']} GB capacity,\n"
            f"    {nb_storage_units} {storage_type} units with {config[storage_type]['capacity']['default']} GB capacity,")

        total_gwp_embedded_value = impact["impacts"]["gwp"]["embedded"]["value"]
        total_gwp_embedded_unit = impact["impacts"]["gwp"]["unit"]

        if nb_storage_units > 0:
            storage_gwp_embedded_value = impact["verbose"][f"{storage_type}-1"]["impacts"]["gwp"]["embedded"]["value"]
            storage_gwp_embedded_unit = impact["verbose"][f"{storage_type}-1"]["impacts"]["gwp"]["unit"]

            assert total_gwp_embedded_unit == storage_gwp_embedded_unit
        else:
            storage_gwp_embedded_value = 0
            storage_gwp_embedded_unit = "kg"

        average_power_value = impact["verbose"]["avg_power"]["value"]
        average_power_unit = impact["verbose"]["avg_power"]["unit"]

        print(
            f"    Impact fabrication compute: {total_gwp_embedded_value - storage_gwp_embedded_value} {total_gwp_embedded_unit},\n"
            f"    Impact fabrication storage: {storage_gwp_embedded_value} {storage_gwp_embedded_unit},\n"
            f"    Average power: {round(average_power_value, 1)} {average_power_unit}\n")


if __name__ == "__main__":
    from time import perf_counter
    start = perf_counter()
    print_archetypes_and_their_configs()
    print(f"Execution time: {perf_counter() - start} seconds")



from abc import abstractmethod
from typing import List

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject


class HardwareBase(ModelingObject):
    # Mark the class as abstract but not its children when they define a default_values class attribute
    @classmethod
    @abstractmethod
    def default_values(cls):
        pass

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity, power: ExplainableQuantity,
                 lifespan: ExplainableQuantity, fraction_of_usage_time: ExplainableQuantity):
        super().__init__(name)
        self.carbon_footprint_fabrication = carbon_footprint_fabrication.set_label(
            f"Carbon footprint fabrication of {self.name}")
        self.power = power.set_label(f"Power of {self.name}")
        self.lifespan = lifespan.set_label(f"Lifespan of {self.name}")
        self.fraction_of_usage_time = fraction_of_usage_time.set_label(f"{self.name} fraction of usage time")

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return self.modeling_obj_containers


class InsufficientCapacityError(Exception):
    def __init__(
            self, overloaded_object: HardwareBase, capacity_type: str,
            available_capacity: ExplainableQuantity|EmptyExplainableObject,
            requested_capacity: ExplainableQuantity|EmptyExplainableObject):
        self.overloaded_object = overloaded_object
        self.capacity_type = capacity_type
        self.available_capacity = available_capacity
        self.requested_capacity = requested_capacity

        message = (f"{self.overloaded_object.name} has available {capacity_type} capacity of "
                   f"{available_capacity.value} but is asked for {requested_capacity.value}")
        super().__init__(message)



from typing import TYPE_CHECKING, List

from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.core.usage.edge.recurrent_edge_device_need import RecurrentEdgeDeviceNeed
from efootprint.core.usage.edge.recurrent_server_need import RecurrentServerNeed

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_usage_journey import EdgeUsageJourney


class EdgeFunction(ModelingObject):
    def __init__(self, name: str, recurrent_edge_device_needs: List[RecurrentEdgeDeviceNeed],
                 recurrent_server_needs: List[RecurrentServerNeed]):
        super().__init__(name)
        self.recurrent_edge_device_needs = recurrent_edge_device_needs
        self.recurrent_server_needs = recurrent_server_needs

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[RecurrentEdgeDeviceNeed]:
        return self.recurrent_edge_device_needs + self.recurrent_server_needs

    @property
    def edge_usage_journeys(self) -> List["EdgeUsageJourney"]:
        return self.modeling_obj_containers



from typing import List, TYPE_CHECKING

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.edge.edge_component import EdgeComponent
from efootprint.core.hardware.hardware_base import InsufficientCapacityError

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern


class EdgeWorkloadComponent(EdgeComponent):
    compatible_root_units = [u.concurrent]
    default_values = {
        "carbon_footprint_fabrication": SourceValue(100 * u.kg),
        "power": SourceValue(50 * u.W),
        "lifespan": SourceValue(6 * u.year),
        "idle_power": SourceValue(5 * u.W),
    }

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity,
                 power: ExplainableQuantity, lifespan: ExplainableQuantity, idle_power: ExplainableQuantity):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan, idle_power)
        self.unitary_hourly_workload_per_usage_pattern = ExplainableObjectDict()

    @property
    def calculated_attributes(self):
        return ["unitary_hourly_workload_per_usage_pattern"] + super().calculated_attributes

    def update_dict_element_in_unitary_hourly_workload_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_hourly_workload = sum(
            [need.unitary_hourly_need_per_usage_pattern[usage_pattern]
             for need in self.recurrent_edge_component_needs if usage_pattern in need.edge_usage_patterns],
            start=EmptyExplainableObject())

        self.unitary_hourly_workload_per_usage_pattern[usage_pattern] = unitary_hourly_workload.set_label(
            f"{self.name} hourly workload for {usage_pattern.name}")

    def update_unitary_hourly_workload_per_usage_pattern(self):
        self.unitary_hourly_workload_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_hourly_workload_per_usage_pattern(usage_pattern)

    def update_dict_element_in_unitary_power_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        if usage_pattern in self.unitary_hourly_workload_per_usage_pattern:
            workload = self.unitary_hourly_workload_per_usage_pattern[usage_pattern]
        else:
            workload = EmptyExplainableObject()

        if isinstance(workload, EmptyExplainableObject):
            unitary_power = self.idle_power
        else:
            unitary_power = self.idle_power + (self.power - self.idle_power) * workload

        self.unitary_power_per_usage_pattern[usage_pattern] = unitary_power.set_label(
            f"{self.name} unitary power for {usage_pattern.name}")

    def update_unitary_power_per_usage_pattern(self):
        self.unitary_power_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_power_per_usage_pattern(usage_pattern)



from efootprint.logger import logger


def rename_dict_key(d, old_key, new_key):
    if old_key not in d:
        raise KeyError(f"{old_key} not found in dictionary")
    if new_key in d:
        raise KeyError(f"{new_key} already exists in dictionary")

    keys = list(d.keys())
    index = keys.index(old_key)
    value = d[old_key]

    # Remove old key
    del d[old_key]

    # Rebuild the dict by inserting the new key at the same position
    d_items = list(d.items())
    d_items.insert(index, (new_key, value))

    d.clear()
    d.update(d_items)


def upgrade_version_9_to_10(system_dict, efootprint_classes_dict=None):
    object_keys_to_delete = ["year", "job_type", "description"]
    for class_key in system_dict:
        if class_key == "efootprint_version":
            continue
        for efootprint_obj_key in system_dict[class_key]:
            for object_key_to_delete in object_keys_to_delete:
                if object_key_to_delete in system_dict[class_key][efootprint_obj_key]:
                    del system_dict[class_key][efootprint_obj_key][object_key_to_delete]
    if "Hardware" in system_dict:
        logger.info(f"Upgrading system dict from version 9 to 10, changing 'Hardware' key to 'Device'")
        system_dict["Device"] = system_dict.pop("Hardware")

    return system_dict


def upgrade_version_10_to_11(system_dict, efootprint_classes_dict=None):
    for system_key in system_dict["System"]:
        system_dict["System"][system_key]["edge_usage_patterns"] = []

    for server_type in ["Server", "GPUServer", "BoaviztaCloudServer"]:
        if server_type not in system_dict:
            continue
        for server_key in system_dict[server_type]:
            rename_dict_key(system_dict[server_type][server_key], "server_utilization_rate", "utilization_rate")

    return system_dict


def upgrade_version_11_to_12(system_dict, efootprint_classes_dict=None):
    if "EdgeDevice" in system_dict:
        system_dict["EdgeComputer"] = system_dict.pop("EdgeDevice")

    if "EdgeUsageJourney" in system_dict:
        logger.info(f"Upgrading system dict from version 11 to 12, upgrading EdgeUsageJourney structure "
                    f"and changing 'EdgeDevice' key to 'EdgeComputer'")
        # Create EdgeFunction entries from edge_processes
        if "EdgeFunction" not in system_dict:
            system_dict["EdgeFunction"] = {}

        for edge_usage_journey_id in system_dict["EdgeUsageJourney"]:
            journey = system_dict["EdgeUsageJourney"][edge_usage_journey_id]

            # Get the edge_device (now edge_computer) reference from the journey
            edge_computer_id = journey.get("edge_device")
            del journey["edge_device"]

            # Embed edge_processes into an edge_function
            edge_function_id = f"ef_{edge_usage_journey_id}"
            edge_process_ids = journey.get("edge_processes", [])
            system_dict["EdgeFunction"][edge_function_id] = {
                "name": f"Edge function for edge usage journey {journey["name"]}",
                "id": edge_function_id,
                "recurrent_edge_resource_needs": edge_process_ids
            }

            # Replace edge_processes with edge_functions
            rename_dict_key(journey, "edge_processes", "edge_functions")
            journey["edge_functions"] = [edge_function_id]

            for edge_process_id in edge_process_ids:
                # Add edge_computer reference to RecurrentEdgeProcess
                system_dict["RecurrentEdgeProcess"][edge_process_id]["edge_device"] = edge_computer_id

    return system_dict


def upgrade_version_12_to_13(system_dict, efootprint_classes_dict=None):
    """
    Upgrade from version 12 to 13: Replace dimensionless units with occurrence/concurrent,
    and byte units with byte_ram where appropriate in timeseries data.
    """
    from efootprint.api_utils.unit_mappings import (
        TIMESERIES_UNIT_MIGRATIONS, SCALAR_RAM_ATTRIBUTES_TO_MIGRATE, RAM_TIMESERIES_ATTRIBUTES_TO_MIGRATE
    )

    logger.info("Upgrading system dict from version 12 to 13: migrating units in timeseries and RAM data")

    def migrate_timeseries_unit(obj_dict, attr_name, new_unit):
        """Migrate unit in timeseries (ExplainableHourlyQuantities or ExplainableRecurrentQuantities) stored in JSON."""
        if attr_name not in obj_dict:
            return

        attr_value = obj_dict[attr_name]

        # Check if it's a timeseries (has 'compressed_values', 'values', or 'recurring_values')
        if isinstance(attr_value, dict) and ('compressed_values' in attr_value or 'values' in attr_value or 'recurring_values' in attr_value):
            if 'unit' in attr_value and attr_value['unit'] in ['dimensionless', '']:
                old_unit = attr_value['unit']
                attr_value['unit'] = new_unit

        # Handle ExplainableObjectDict (dict of timeseries)
        elif isinstance(attr_value, dict):
            for key, sub_value in attr_value.items():
                if isinstance(sub_value, dict) and ('compressed_values' in sub_value or 'values' in sub_value or 'recurring_values' in sub_value):
                    if 'unit' in sub_value and sub_value['unit'] in ['dimensionless', '']:
                        old_unit = sub_value['unit']
                        sub_value['unit'] = new_unit

    def migrate_ram_timeseries_unit(obj_dict, attr_name):
        """Migrate unit in RAM timeseries (ExplainableHourlyQuantities or ExplainableRecurrentQuantities) by appending _ram."""
        if attr_name not in obj_dict:
            return

        attr_value = obj_dict[attr_name]

        # Check if it's a timeseries (has 'compressed_values', 'values', or 'recurring_values')
        if isinstance(attr_value, dict) and ('compressed_values' in attr_value or 'values' in attr_value or 'recurring_values' in attr_value):
            if 'unit' in attr_value:
                old_unit = attr_value['unit']
                # Only migrate if it's a byte unit (not already _ram)
                if '_ram' not in old_unit and any(byte_prefix in old_unit.lower() for byte_prefix in ['byte', 'b']):
                    # Append _ram to the existing unit to preserve power of ten
                    new_unit = old_unit + '_ram' if old_unit.endswith('byte') else old_unit.replace('B', 'B_ram')
                    attr_value['unit'] = new_unit

    def migrate_scalar_ram_unit(obj_dict, attr_name):
        """Migrate unit in scalar ExplainableQuantity stored in JSON by appending _ram."""
        if attr_name not in obj_dict:
            return

        attr_value = obj_dict[attr_name]

        # Check if it's a scalar ExplainableQuantity (has 'unit' but not timeseries keys)
        if isinstance(attr_value, dict) and 'unit' in attr_value:
            if 'compressed_values' not in attr_value and 'values' not in attr_value and 'recurring_values' not in attr_value:
                old_unit = attr_value['unit']
                # Only migrate if it's a byte unit (not already _ram)
                if '_ram' not in old_unit and any(byte_prefix in old_unit.lower() for byte_prefix in ['byte', 'b']):
                    # Append _ram to the existing unit to preserve power of ten
                    new_unit = old_unit + '_ram' if old_unit.endswith('byte') else old_unit.replace('B', 'B_ram')
                    attr_value['unit'] = new_unit

    # Iterate through all classes and objects
    for class_name in system_dict:
        if class_name == "efootprint_version":
            continue
        efootprint_class = efootprint_classes_dict[class_name]

        for obj_id in system_dict[class_name]:
            obj_dict = system_dict[class_name][obj_id]

            # Apply timeseries unit migrations (dimensionless -> occurrence/concurrent)
            for (migration_class, attr_name), new_unit in TIMESERIES_UNIT_MIGRATIONS.items():
                if efootprint_class.is_subclass_of(migration_class):
                    migrate_timeseries_unit(obj_dict, attr_name, new_unit)

            # Apply RAM timeseries unit migrations (append _ram)
            for (migration_class, attr_name) in RAM_TIMESERIES_ATTRIBUTES_TO_MIGRATE:
                if efootprint_class.is_subclass_of(migration_class):
                    migrate_ram_timeseries_unit(obj_dict, attr_name)

            # Apply scalar RAM unit migrations (append _ram)
            for (migration_class, attr_name) in SCALAR_RAM_ATTRIBUTES_TO_MIGRATE:
                if efootprint_class.is_subclass_of(migration_class):
                    migrate_scalar_ram_unit(obj_dict, attr_name)

    return system_dict


def upgrade_version_13_to_14(system_dict, efootprint_classes_dict=None):
    if "EdgeComputer" in system_dict:
        for edge_computer_id in system_dict["EdgeComputer"]:
            del system_dict["EdgeComputer"][edge_computer_id]["power_usage_effectiveness"]
            del system_dict["EdgeComputer"][edge_computer_id]["utilization_rate"]
            system_dict["EdgeComputer"][edge_computer_id]["structure_carbon_footprint_fabrication"] = \
                system_dict["EdgeComputer"][edge_computer_id]["carbon_footprint_fabrication"]
    if "EdgeFunction" in system_dict:
        logger.info("Upgrading system dict from version 13 to 14: renaming recurrent_edge_resource_needs to "
                    "recurrent_edge_device_needs in EdgeFunctions and updating EdgeComputer attributes")
        for edge_function_id in system_dict["EdgeFunction"]:
            rename_dict_key(system_dict["EdgeFunction"][edge_function_id], "recurrent_edge_resource_needs",
                            "recurrent_edge_device_needs")

    return system_dict


def upgrade_version_14_to_15(system_dict, efootprint_classes_dict=None):
    if "EdgeUsagePattern" in system_dict:
        logger.info("Upgrading system dict from version 14 to 15: adding default wifi network to EdgeUsagePatterns"
                    " and empty recurrent_server_needs to EdgeFunctions")
        default_network_id = "default_wifi_network_for_edge"
        if "Network" not in system_dict:
            system_dict["Network"] = {}
        system_dict["Network"][default_network_id] = {
            "name": "Default wifi network for edge",
            "id": default_network_id,
            "bandwidth_energy_intensity": {
                "value": 0.05, "unit": "kilowatt_hour / gigabyte",
                "label": "bandwith energy intensity of Default wifi network from e-footprint hypothesis",
                "source": {"name": "e-footprint hypothesis", "link": None}
            }
        }
        for edge_usage_pattern_id in system_dict["EdgeUsagePattern"]:
            system_dict["EdgeUsagePattern"][edge_usage_pattern_id]["network"] = default_network_id

    if "EdgeFunction" in system_dict:
        for edge_function_id in system_dict["EdgeFunction"]:
            system_dict["EdgeFunction"][edge_function_id]["recurrent_server_needs"] = []

    return system_dict


VERSION_UPGRADE_HANDLERS = {
    9: upgrade_version_9_to_10,
    10: upgrade_version_10_to_11,
    11: upgrade_version_11_to_12,
    12: upgrade_version_12_to_13,
    13: upgrade_version_13_to_14,
    14: upgrade_version_14_to_15,
}



"""Shared utilities for timeseries aggregation strategies."""

from pint import Unit, Quantity


def get_plot_aggregation_strategy(unit: Unit) -> str:
    """
    Determine how to aggregate hourly values into daily values for plotting.

    Args:
        unit: The Pint unit of the timeseries

    Returns:
        'sum' for cumulative metrics (events, energy, data transfer)
        'mean' for concurrent/resource allocation metrics (instances, RAM, CPU)
    """
    unit_str = str(unit)

    # Rate and resource allocation units → mean
    if 'concurrent' in unit_str or '_ram' in unit_str or 'cpu_core' in unit_str or 'gpu' in unit_str:
        return 'mean'

    # Default: sum for events, energy, mass, data transfer
    return 'sum'


def validate_timeseries_unit(value: Quantity, label: str = None):
    """
    Validate that a timeseries value does not use dimensionless unit.

    Args:
        value: The Pint Quantity to validate
        label: Optional label for error message

    Raises:
        ValueError: If the value uses dimensionless unit
    """
    from efootprint.constants.units import u

    # Allow dimensionless for unlabeled intermediate calculations (e.g., during division operations)
    # These will be converted to occurrence/concurrent before being assigned
    if value.units == u.dimensionless and label:
        raise ValueError(
            f"Timeseries cannot use dimensionless unit. "
            f"Use 'occurrence' (for discrete occurrences, aggregated by sum) or "
            f"'concurrent' (for concurrent counts, aggregated by mean). "
            f"Label: {label}"
        )


from typing import List

import pytz

from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.explainable_timezone import ExplainableTimezone
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceObject
from efootprint.constants.units import u


class Country(ModelingObject):
    default_values =  {
            "average_carbon_intensity": SourceValue(50 * u.g / u.kWh, label="Average carbon intensity of the country"),
            "timezone": SourceObject(pytz.timezone('Europe/Paris'), label="Country timezone")
        }

    def __init__(
            self, name: str, short_name: str, average_carbon_intensity: ExplainableQuantity,
            timezone: ExplainableTimezone):
        super().__init__(name)
        self.short_name = short_name
        self.average_carbon_intensity = average_carbon_intensity.set_label(f"Average carbon intensity of {self.name}")
        self.timezone = timezone.set_label(f"{self.name} timezone")

    @property
    def attributes_that_shouldnt_trigger_update_logic(self):
        return super().attributes_that_shouldnt_trigger_update_logic + ["short_name"]

    @property
    def usage_patterns(self):
        return self.modeling_obj_containers

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self):
        return self.usage_patterns



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.18.1
#   kernelspec:
#     display_name: Jupyter AI
#     language: python
#     name: jupyter-ai
# ---

# %%
import pandas as pd

df = pd.read_csv("carbon-intensity-electricity.csv")
pop = pd.read_csv("world_population.csv")
pop = pop.rename(columns={'Country Code': 'Code', '2024': 'Population'})
df = df.merge(pop[['Code', 'Population']], on='Code', how='left')
# Keep only last year for each country
df = df.sort_values(["Entity", "Year"])
df = df.groupby("Entity").tail(1)
df_without_code = df[pd.isnull(df["Code"])]
df = df[~pd.isnull(df["Code"])]
df.head()

# %%
df_without_code.head()

# %%
# Entities with no population
df[pd.isnull(df["Population"])]

# %%
big_countries = df[df["Population"]>5000000].sort_values(by="Entity").copy()
big_countries

# %%
import pycountry
import pytz

def get_alpha2(country):
    """
    country: either alpha-3 code (e.g. 'FRA'), or full name ('France'),
             or short name ('United States'), etc.
    Returns: ISO alpha-2 code (e.g. 'FR')
    """
    # If this looks like an alpha-3 code
    if len(country) == 3 and country.isalpha():
        try:
            c = pycountry.countries.get(alpha_3=country.upper())
            if c:
                return c.alpha_2
        except KeyError:
            pass

    # Try exact name match
    try:
        c = pycountry.countries.lookup(country)
        return c.alpha_2
    except LookupError:
        return None



# %%
def country_timezones(country):
    alpha2 = get_alpha2(country)
    if not alpha2:
        raise ValueError(f"No alpha2 found for {country}")

    tzens = pytz.country_timezones.get(alpha2)
    if not tzens:
        raise ValueError(f"No tz found for {country}")

    return tzens



# %%
countries_with_multiple_timezones = {}
for index, row in big_countries.iterrows():
    tzens = country_timezones(row["Code"])
    if len(tzens) > 1:
        countries_with_multiple_timezones[row["Code"]] = tzens

# %%
capitals_timezones_dict = {
    "ARG": "America/Argentina/Buenos_Aires",
    "AUS": "Australia/Sydney",
    "BRA": "America/Sao_Paulo",
    "CAN": "America/Toronto",
    "CHL": "America/Santiago",
    "CHN": "Asia/Shanghai",
    "COD": "Africa/Kinshasa",
    "ECU": "America/Guayaquil",
    "DEU": "Europe/Berlin",
    "IDN": "Asia/Jakarta",
    "KAZ": "Asia/Almaty",
    "MYS": "Asia/Kuala_Lumpur",
    "MEX": "America/Mexico_City",
    "PNG": "Pacific/Port_Moresby",
    "PRT": "Europe/Lisbon",
    "RUS": "Europe/Moscow",
    "ESP": "Europe/Madrid",
    "UKR": "Europe/Kyiv",
    "USA": "America/New_York",
    "UZB": "Asia/Tashkent"
}

# %%
for capital, timezone in capitals_timezones_dict.items():
    assert timezone in countries_with_multiple_timezones[capital]


# %%
def country_timezone(country_code: str):
    if country_code in capitals_timezones_dict:
        return capitals_timezones_dict[country_code]
    else:
        return country_timezones(country_code)[0]


# %%
big_countries["Timezone"] = big_countries["Code"].apply(country_timezone)

# %%
big_countries

# %%
assert not big_countries.isnull().values.any(), "Null values detected in big_countries"

# %%
big_countries.drop("Population", axis=1).to_csv("countries_elec_carbon_intensity_and_timezone.csv", index=False)

# %%
# automatically generate Countries class attributes
for country in big_countries["Entity"]:
    print(f"{country.replace(" ", "_").upper()} = country_generator_from_csv(\"{country}\")")

# %%



from datetime import datetime, timedelta
from typing import Optional

import numpy as np
from pint import Quantity


def get_time_axis(start_date: datetime, length: int) -> np.ndarray:
    return np.array([start_date + timedelta(hours=i) for i in range(length)])


def prepare_data(q: Quantity, start: datetime, apply_cumsum=False) -> tuple[np.ndarray, Quantity]:
    magnitudes = np.cumsum(q.magnitude) if apply_cumsum else q.magnitude
    return get_time_axis(start, len(magnitudes)), Quantity(magnitudes, q.units)


def plot_baseline_and_simulation_data(
    baseline_q: Quantity, time_baseline: np.ndarray, simulated_q: Optional[Quantity] = None,
    time_sim: Optional[np.ndarray] = None, figsize=(10, 4), xlims: Optional[tuple[datetime, datetime]] = None):
    from matplotlib import pyplot as plt
    import matplotlib.dates as mdates

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

    if simulated_q is not None:
        simulated_q = simulated_q.to(baseline_q.units)

    if xlims is not None:
        start, end = xlims
        baseline_mask = (time_baseline >= start) & (time_baseline <= end)
        time_baseline_plot = time_baseline[baseline_mask]
        baseline_plot_q = baseline_q[baseline_mask]

        if simulated_q is not None and time_sim is not None:
            sim_mask = (time_sim >= start) & (time_sim <= end)
            time_sim_plot = time_sim[sim_mask]
            simulated_plot_q = simulated_q[sim_mask]
        else:
            time_sim_plot = None
            simulated_plot_q = None

        # Y-axis autoscaling with margin
        all_vals = [baseline_plot_q.magnitude]
        if simulated_q is not None:
            all_vals.append(simulated_plot_q.magnitude)

        min_val = np.min([v.min() for v in all_vals])
        max_val = np.max([v.max() for v in all_vals])
        offset = (max_val - min_val) * 0.1 if max_val != min_val else 1
        ax.set_ylim(min_val - offset, max_val + offset)
        ax.set_xlim(start, end)
    else:
        # Use full data
        time_baseline_plot = time_baseline
        baseline_plot_q = baseline_q
        time_sim_plot = time_sim
        simulated_plot_q = simulated_q

    ax.plot(time_baseline_plot, baseline_plot_q.magnitude, label="baseline")

    if simulated_q is not None and time_sim_plot is not None:
        ax.plot(time_sim_plot, simulated_plot_q.magnitude, label="simulated")

    ax.legend()
    plt.ylabel(f"{baseline_q.units:~}")

    locator = mdates.AutoDateLocator(minticks=3, maxticks=12)
    formatter = mdates.ConciseDateFormatter(locator)
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)

    return fig, ax



from datetime import datetime

from pint import Quantity

from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, Source
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.explainable_recurrent_quantities import ExplainableRecurrentQuantities
from efootprint.abstract_modeling_classes.explainable_timezone import ExplainableTimezone
from efootprint.constants.sources import Sources


SOURCE_VALUE_DEFAULT_NAME = "unnamed source"


class SourceObject(ExplainableObject):
    __slots__ = ()

    def __init__(self, value: object, source: Source = Sources.HYPOTHESIS, label: str = SOURCE_VALUE_DEFAULT_NAME):
        super().__init__(value, label=label, source=source)

class SourceTimezone(ExplainableTimezone):
    __slots__ = ()

    def __init__(self, value: object, source: Source = Sources.HYPOTHESIS, label: str = SOURCE_VALUE_DEFAULT_NAME):
        super().__init__(value, label=label, source=source)

class SourceValue(ExplainableQuantity):
    __slots__ = ()

    def __init__(self, value: Quantity, source: Source = Sources.HYPOTHESIS, label: str = SOURCE_VALUE_DEFAULT_NAME):
        super().__init__(value, label=label, source=source)

class SourceHourlyValues(ExplainableHourlyQuantities):
    __slots__ = ()

    def __init__(self, value: Quantity, start_date: datetime, source: Source = Sources.HYPOTHESIS,
                 label: str = SOURCE_VALUE_DEFAULT_NAME):
        super().__init__(value, start_date=start_date, label=label, source=source)

class SourceRecurrentValues(ExplainableRecurrentQuantities):
    __slots__ = ()

    def __init__(self, value: Quantity, source: Source = Sources.HYPOTHESIS, label: str = SOURCE_VALUE_DEFAULT_NAME):
        super().__init__(value, label=label, source=source)



from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.hardware_base import HardwareBase


class Device(HardwareBase):
    default_values =  {
            "carbon_footprint_fabrication": SourceValue(150 * u.kg),
            "power": SourceValue(50 * u.W),
            "lifespan": SourceValue(6 * u.year),
            "fraction_of_usage_time": SourceValue(7 * u.hour / u.day)
        }

    @classmethod
    def smartphone(cls, name="Default smartphone", **kwargs):
        output_args = {
            "carbon_footprint_fabrication": SourceValue(30 * u.kg, Sources.BASE_ADEME_V19),
            "power": SourceValue(1 * u.W),
            "lifespan": SourceValue(3 * u.year),
            "fraction_of_usage_time": SourceValue(3.6 * u.hour / u.day, Sources.STATE_OF_MOBILE_2022)
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def laptop(cls, name="Default laptop", **kwargs):
        output_args = {
            "carbon_footprint_fabrication": SourceValue(156 * u.kg, Sources.BASE_ADEME_V19),
            "power": SourceValue(50 * u.W),
            "lifespan": SourceValue(6 * u.year),
            "fraction_of_usage_time": SourceValue(7 * u.hour / u.day, Sources.STATE_OF_MOBILE_2022)
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def box(cls, name="Default box", **kwargs):
        output_args = {
            "carbon_footprint_fabrication": SourceValue(78 * u.kg, Sources.BASE_ADEME_V19),
            "power": SourceValue(10 * u.W),
            "lifespan": SourceValue(6 * u.year),
            "fraction_of_usage_time": SourceValue(24 * u.hour / u.day)
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def screen(cls, name="Default screen", **kwargs):
        output_args = {
            "carbon_footprint_fabrication": SourceValue(222 * u.kg, Sources.BASE_ADEME_V19),
            "power": SourceValue(30 * u.W),
            "lifespan": SourceValue(6 * u.year),
            "fraction_of_usage_time": SourceValue(7 * u.hour / u.day)
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def archetypes(cls):
        return [cls.smartphone, cls.laptop, cls.box, cls.screen]

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity, power: ExplainableQuantity,
                 lifespan: ExplainableQuantity, fraction_of_usage_time: ExplainableQuantity):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan, fraction_of_usage_time)



from typing import List, TYPE_CHECKING

import numpy as np

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.core.hardware.edge.edge_device import EdgeDevice
from efootprint.abstract_modeling_classes.explainable_recurrent_quantities import ExplainableRecurrentQuantities
from efootprint.core.usage.job import JobBase
from efootprint.constants.units import u

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_function import EdgeFunction
    from efootprint.core.usage.edge.edge_usage_journey import EdgeUsageJourney
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern
    

class NegativeServerNeedError(Exception):
    def __init__(self, server_need_name: str, min_value: float):
        message = (
            f"Server need '{server_need_name}' has negative values (min {min_value:.3f}). "
            f"Server need volumes must be positive.")
        super().__init__(message)


class RecurrentServerNeed(ModelingObject):
    default_values = {
        "recurrent_volume_per_edge_device": ExplainableRecurrentQuantities(
            np.array([1.0] * 168, dtype=np.float32) * u.occurrence, label="Default recurrent volume per edge device")
    }

    def __init__(self, name: str, edge_device: EdgeDevice,
                 recurrent_volume_per_edge_device: ExplainableRecurrentQuantities,
                 jobs: List[JobBase]):
        super().__init__(name)
        self.edge_device = edge_device
        self.recurrent_volume_per_edge_device = recurrent_volume_per_edge_device
        self.jobs = jobs
        
        self.validated_recurrent_need = EmptyExplainableObject()
        self.unitary_hourly_volume_per_usage_pattern = ExplainableObjectDict()

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[JobBase]:
        return self.jobs

    @property
    def edge_functions(self) -> List["EdgeFunction"]:
        return self.modeling_obj_containers

    @property
    def edge_usage_journeys(self) -> List["EdgeUsageJourney"]:
        return list(set(sum([ef.edge_usage_journeys for ef in self.edge_functions], start=[])))

    @property
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        return list(set(sum([euj.edge_usage_patterns for euj in self.edge_usage_journeys], start=[])))

    @property
    def calculated_attributes(self):
        return ["validated_recurrent_need", "unitary_hourly_volume_per_usage_pattern"]

    def update_validated_recurrent_need(self):
        assert self.recurrent_volume_per_edge_device.unit == u.occurrence, \
            (f"RecurrentServerNeed '{self.name}' has invalid unit '{self.recurrent_volume_per_edge_device.unit}', "
             f"expected 'occurrence'.")
        min_value = np.min(self.recurrent_volume_per_edge_device.magnitude)
        if min_value < 0:
            raise NegativeServerNeedError(self.name, min_value)
        self.validated_recurrent_need = self.recurrent_volume_per_edge_device.copy().set_label(
            f"{self.name} validated recurrent need")
        
    def update_dict_element_in_unitary_hourly_volume_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_hourly_volume = self.recurrent_volume_per_edge_device.generate_hourly_quantities_over_timespan(
            usage_pattern.nb_edge_usage_journeys_in_parallel, usage_pattern.country.timezone)
        self.unitary_hourly_volume_per_usage_pattern[usage_pattern] = unitary_hourly_volume.set_label(
            f"{self.name} unitary hourly need for {usage_pattern.name}")

    def update_unitary_hourly_volume_per_usage_pattern(self):
        self.unitary_hourly_volume_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_hourly_volume_per_usage_pattern(usage_pattern)



from collections import deque
from copy import copy
from typing import Type, Optional, TYPE_CHECKING
from dataclasses import dataclass
import os

from efootprint.abstract_modeling_classes.object_linked_to_modeling_obj import ObjectLinkedToModelingObj
from efootprint.constants.units import u
from efootprint.logger import logger
from efootprint.utils.calculus_graph import build_calculus_graph
from efootprint.utils.graph_tools import add_unique_id_to_mynetwork

if TYPE_CHECKING:
    from efootprint.abstract_modeling_classes.modeling_object import ModelingObject


@dataclass
class Source:
    @classmethod
    def from_json_dict(cls, d):
        if "name" not in d:
            raise ValueError("Source JSON must contain 'name' field")
        return cls(name=d["name"], link=d.get("link", None))
    name: str
    link: Optional[str]

    def to_json(self):
        output_dict = {"name": self.name}
        if self.link is not None:
            output_dict["link"] = self.link
        return output_dict


def retrieve_update_function_from_mod_obj_and_attr_name(mod_obj: "ModelingObject", attr_name: str):
    update_func_name = f"update_{attr_name}"
    update_func = getattr(mod_obj, update_func_name, None)

    if update_func is None:
        raise AttributeError(f"No update function associated to {attr_name} in {mod_obj.class_as_simple_str} "
                             f"{mod_obj.name} ({mod_obj.id}), please create it.")

    return update_func


def retrieve_dict_element_update_function_from_mod_obj_and_attr_name(
        mod_obj: "ModelingObject", attr_name: str, key_in_dict: "ModelingObject"):
    update_func_name = f"update_dict_element_in_{attr_name}"
    update_func = getattr(mod_obj, update_func_name, None)

    if update_func is None:
        raise AttributeError(
            f"No dict element update function associated to {attr_name} in {mod_obj.id}, please create it.")

    return update_func


def optimize_attr_updates_chain(attr_updates_chain):
    initial_chain_len = len(attr_updates_chain)
    attr_to_update_ids = [attr.id for attr in attr_updates_chain]
    optimized_chain = []

    for index in range(len(attr_updates_chain)):
        attr_to_update = attr_updates_chain[index]
        is_not_recomputed_later = attr_to_update.id not in attr_to_update_ids[index + 1:]
        doesnt_belong_to_dict = attr_to_update.dict_container is None
        belongs_to_dict_and_dict_is_not_recomputed_later = (
            attr_to_update.dict_container is not None and
            attr_to_update.dict_container.id not in attr_to_update_ids[index + 1:]
        )

        if is_not_recomputed_later and (doesnt_belong_to_dict or belongs_to_dict_and_dict_is_not_recomputed_later):
            # Keep only last occurrence of each attribute to update
            optimized_chain.append(attr_to_update)

    optimized_chain_len = len(optimized_chain)

    if optimized_chain_len != initial_chain_len:
        logger.info(f"Optimized update function chain from {initial_chain_len} to {optimized_chain_len} calculations")

    return optimized_chain


def get_attribute_from_flat_obj_dict(attr_key: str, flat_obj_dict: dict):
    modeling_obj_container_id, attr_name_in_mod_obj_container, key_in_dict = eval(attr_key)
    if key_in_dict:
        return getattr(flat_obj_dict[modeling_obj_container_id], attr_name_in_mod_obj_container)[
            flat_obj_dict[key_in_dict]]
    else:
        return getattr(flat_obj_dict[modeling_obj_container_id], attr_name_in_mod_obj_container)


class ExplainableObject(ObjectLinkedToModelingObj):
    __slots__ = (
        # ExplainableObject's own attributes (parent ObjectLinkedToModelingObj has its own slots)
        'simulation_twin',
        'baseline_twin',
        'simulation',
        'initial_modeling_obj_container',
        '_value',
        'source',
        'label',
        'left_parent',
        'right_parent',
        'operator',
        '_keys_of_direct_ancestors_with_id_loaded_from_json',
        '_keys_of_direct_children_with_id_loaded_from_json',
        'flat_obj_dict',
        '_direct_ancestors_with_id',
        '_direct_children_with_id',
        'explain_nested_tuples_from_json',
        '_explain_nested_tuples',
    )
    _registry = []

    @classmethod
    def register_subclass(cls, matcher):
        def decorator(subclass):
            cls._registry.append((matcher, subclass))
            return subclass

        return decorator

    @classmethod
    def from_json_dict(cls, d):
        for matcher, subclass in cls._registry:
            if matcher(d):
                return subclass.from_json_dict(d)
        if "value" in d and isinstance(d["value"], str):
            source = Source.from_json_dict(d.get("source")) if d.get("source") else None
            return cls(d["value"], label=d.get("label", None), source=source)
        raise ValueError("No matching subclass found for data: {}".format(d))

    def initialize_calculus_graph_data_from_json(self, json_input: dict, flat_obj_dict: dict[str, "ModelingObject"]):
        if "direct_ancestors_with_id" in json_input:
            self._keys_of_direct_ancestors_with_id_loaded_from_json = json_input[
                "direct_ancestors_with_id"]
            self._keys_of_direct_children_with_id_loaded_from_json = json_input[
                "direct_children_with_id"]
            self.explain_nested_tuples_from_json = json_input["explain_nested_tuples"]
            self.flat_obj_dict = flat_obj_dict

    def __init__(
            self, value: object, label: str = None, left_parent: Type["ExplainableObject"] = None,
            right_parent: Type["ExplainableObject"] = None, operator: str = None, source: Source = None):
        super().__init__()
        self.simulation_twin = None
        self.baseline_twin = None
        self.simulation = None
        self.initial_modeling_obj_container = None
        self._value = value
        if not label and (left_parent is None and right_parent is None):
            raise ValueError(f"ExplainableObject without parent should have a label")
        self.source = source
        self.label = None
        self.set_label(label)
        self.left_parent = left_parent
        self.right_parent = right_parent
        self.operator = operator
        self._keys_of_direct_ancestors_with_id_loaded_from_json = None
        self._keys_of_direct_children_with_id_loaded_from_json = None
        self.flat_obj_dict = None
        self._direct_ancestors_with_id = []
        self._direct_children_with_id = []
        self.explain_nested_tuples_from_json = None
        self._explain_nested_tuples = None

        for parent in (self.left_parent, self.right_parent):
            if parent is not None:
                self.direct_ancestors_with_id += [
                    ancestor_with_id for ancestor_with_id in parent.return_direct_ancestors_with_id_to_child()
                    if ancestor_with_id.id not in self.direct_ancestor_ids]


    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value

    @value.deleter
    def value(self):
        self._value = None

    @property
    def explain_nested_tuples(self):
        if self._explain_nested_tuples is None and self.explain_nested_tuples_from_json is not None:
            def recursively_deserialize_explain_nested_tuple(explain_nested_tuple):
                if isinstance(explain_nested_tuple, list) or isinstance(explain_nested_tuple, tuple):
                    return (recursively_deserialize_explain_nested_tuple(explain_nested_tuple[0]),
                            explain_nested_tuple[1],
                            recursively_deserialize_explain_nested_tuple(explain_nested_tuple[2]))
                elif explain_nested_tuple is not None:
                    if isinstance(explain_nested_tuple, str):
                        return get_attribute_from_flat_obj_dict(explain_nested_tuple, self.flat_obj_dict)
                    elif isinstance(explain_nested_tuple, dict):
                        return ExplainableObject.from_json_dict(explain_nested_tuple)
                return None

            self._explain_nested_tuples = recursively_deserialize_explain_nested_tuple(
                self.explain_nested_tuples_from_json)

        return self._explain_nested_tuples

    @explain_nested_tuples.setter
    def explain_nested_tuples(self, new_explain_nested_tuples):
        self._explain_nested_tuples = new_explain_nested_tuples
        self.explain_nested_tuples_from_json = None

    @explain_nested_tuples.deleter
    def explain_nested_tuples(self):
        self._explain_nested_tuples = None
        self.explain_nested_tuples_from_json = None

    def load_ancestors_and_children_from_json(self):
        if (self._keys_of_direct_ancestors_with_id_loaded_from_json is not None
                and self._keys_of_direct_children_with_id_loaded_from_json is not None):
            self.direct_ancestors_with_id = [
                get_attribute_from_flat_obj_dict(direct_ancestor_key, self.flat_obj_dict) for direct_ancestor_key in
                self._keys_of_direct_ancestors_with_id_loaded_from_json
            ]
            self.direct_children_with_id = [
                get_attribute_from_flat_obj_dict(direct_child_key, self.flat_obj_dict) for direct_child_key in
                self._keys_of_direct_children_with_id_loaded_from_json
            ]

    @property
    def direct_ancestors_with_id(self):
        self.load_ancestors_and_children_from_json()

        return self._direct_ancestors_with_id

    @direct_ancestors_with_id.setter
    def direct_ancestors_with_id(self, value):
        self._direct_ancestors_with_id = value
        self._keys_of_direct_ancestors_with_id_loaded_from_json = None

    @property
    def direct_children_with_id(self):
        self.load_ancestors_and_children_from_json()

        return self._direct_children_with_id

    @direct_children_with_id.setter
    def direct_children_with_id(self, value):
        self._direct_children_with_id = value
        self._keys_of_direct_children_with_id_loaded_from_json = None

    def __copy__(self):
        cls = self.__class__
        new_instance = cls.__new__(cls)
        new_instance.__init__(
            value=copy(self.value), label=copy(self.label), source=copy(getattr(self, "source", None)))

        return new_instance

    def set_label(self, new_label):
        if self.source is not None and f"from {self.source.name}" not in new_label:
            self.label = f"{new_label} from {self.source.name}"
        else:
            self.label = new_label

        return self

    @property
    def has_parent(self):
        return self.left_parent is not None or self.right_parent is not None

    @property
    def direct_ancestor_ids(self):
        return [attr.id for attr in self.direct_ancestors_with_id]

    @property
    def direct_child_ids(self):
        return [attr.id for attr in self.direct_children_with_id]

    def replace_in_mod_obj_container_without_recomputation(self, new_value):
        value_to_set = new_value
        if isinstance(value_to_set, float) and self.dict_container is not None:
            from efootprint.abstract_modeling_classes.source_objects import SourceValue
            value_to_set = SourceValue(value_to_set * u.dimensionless)

        super().replace_in_mod_obj_container_without_recomputation(value_to_set)

    def set_modeling_obj_container(
            self, new_modeling_obj_container: Type["ModelingObject"] | None, attr_name: str | None):
        if not self.label:
            raise PermissionError(
                f"ExplainableObjects that are attributes of a ModelingObject should always have a label. "
                f"{self} doesn’t have one.")

        if self.modeling_obj_container is not None:
            for direct_ancestor_with_id in self.direct_ancestors_with_id:
                direct_ancestor_with_id.remove_child_from_direct_children_with_id(direct_child=self)
            for child in self.direct_children_with_id:
                # Rehydrate child calculus graph so that it now has live links to its ancestors
                child.load_ancestors_and_children_from_json()

        super().set_modeling_obj_container(new_modeling_obj_container, attr_name)

        if new_modeling_obj_container is not None:
            if self.initial_modeling_obj_container is None:
                self.initial_modeling_obj_container = new_modeling_obj_container
            for direct_ancestor_with_id in self.direct_ancestors_with_id:
                direct_ancestor_with_id.add_child_to_direct_children_with_id(direct_child=self)

            if self.left_parent is not None or self.right_parent is not None:
                self.explain_nested_tuples = self.compute_explain_nested_tuples()
                # Free up memory because left parent and right_parent aren’t needed anymore
                self.left_parent = None
                self.right_parent = None

    def return_direct_ancestors_with_id_to_child(self):
        if self.modeling_obj_container is not None:
            return [self]
        else:
            return [ancestor for ancestor in self.direct_ancestors_with_id
                    if ancestor.modeling_obj_container is not None]

    def add_child_to_direct_children_with_id(self, direct_child):
        if direct_child.id not in self.direct_child_ids:
            self.direct_children_with_id.append(direct_child)

    def remove_child_from_direct_children_with_id(self, direct_child):
        self.direct_children_with_id = [
            child for child in self.direct_children_with_id if child.id != direct_child.id]

    @property
    def all_descendants_with_id(self):
        all_descendants = []
        visited_ids = set()
        stack = [self]

        while stack:
            parent = stack.pop()
            for child in parent.direct_children_with_id:
                cid = child.id
                if cid not in visited_ids:
                    visited_ids.add(cid)
                    all_descendants.append(child)
                    stack.append(child)

        return all_descendants

    @property
    def all_ancestors_with_id(self):
        all_ancestors = []

        def retrieve_ancestors(expl_obj: ExplainableObject, ancestors_list):
            for parent in expl_obj.direct_ancestors_with_id:
                if parent.id not in [elt.id for elt in ancestors_list]:
                    ancestors_list.append(parent)
                retrieve_ancestors(parent, ancestors_list)

        retrieve_ancestors(self, all_ancestors)

        return all_ancestors

    @property
    def attr_updates_chain(self):
        attr_updates_chain = []
        descendants = self.all_descendants_with_id
        descendant_ids = {desc.id for desc in descendants if desc.id != self.id}
        has_been_added_to_chain_dict = {desc.id: False for desc in descendants if desc.id != self.id}

        # Use deque for efficient pops and removals
        parents_with_children_to_add = deque([self])

        # Precompute ancestor ids for each child
        ancestor_ids_map = {
            child.id: [ancestor.id for ancestor in child.direct_ancestors_with_id]
            for child in descendants
        }

        while parents_with_children_to_add:
            next_parents = deque()
            while parents_with_children_to_add:
                parent = parents_with_children_to_add.popleft()
                keep_for_next_iteration = False

                for child in parent.direct_children_with_id:
                    if not has_been_added_to_chain_dict[child.id]:
                        child_ancestor_ids = ancestor_ids_map.get(child.id, [])
                        all_child_ancestors_that_need_to_be_updated_are_already_in_chain = all(
                            has_been_added_to_chain_dict[ancestor_id] for ancestor_id in child_ancestor_ids if
                               ancestor_id in descendant_ids)
                        if all_child_ancestors_that_need_to_be_updated_are_already_in_chain:
                            attr_updates_chain.append(child)
                            has_been_added_to_chain_dict[child.id] = True

                            if child.direct_children_with_id:
                                next_parents.append(child)
                        else:
                            keep_for_next_iteration = True

                if keep_for_next_iteration:
                    next_parents.append(parent)

            parents_with_children_to_add = next_parents

        optimized_chain = optimize_attr_updates_chain(attr_updates_chain)
        return optimized_chain

    @property
    def update_function(self):
        if self.modeling_obj_container is None:
            raise ValueError(
                f"{self} doesn’t have a modeling_obj_container, hence it makes no sense "
                f"to look for its update function")
        dict_container = self.dict_container
        if dict_container is None:
            update_func = retrieve_update_function_from_mod_obj_and_attr_name(
                self.modeling_obj_container, self.attr_name_in_mod_obj_container)
        else:
            update_func = retrieve_dict_element_update_function_from_mod_obj_and_attr_name(
                self.modeling_obj_container, self.attr_name_in_mod_obj_container, self.key_in_dict)

        return update_func

    @property
    def update_function_chain(self):
        return [attribute.update_function for attribute in self.attr_updates_chain]
    
    def generate_explainable_object_with_logical_dependency(
            self, explainable_condition: Type["ExplainableObject"]):
        return self.__class__(value=self.value, label=self.label, left_parent=self, right_parent=explainable_condition,
                              operator="logically dependent on")

    def explain(self, pretty_print=True):
        element_value_to_print = str(self)

        if len(self.direct_ancestors_with_id) == 0:
            return f"{self.label} = {element_value_to_print}"

        flat_tuple_formula = self.compute_formula_as_flat_tuple(self.explain_nested_tuples)

        if pretty_print:
            return self.pretty_print_calculation(
                f"{self.label} = {self.print_flat_tuple_formula(flat_tuple_formula, print_values_instead_of_labels=False)}"
                f" = {self.print_flat_tuple_formula(flat_tuple_formula, print_values_instead_of_labels=True)}"
                f" = {element_value_to_print}")
        else:
            return f"{self.label} = {self.print_flat_tuple_formula(
                flat_tuple_formula, print_values_instead_of_labels=False)}" \
                f" = {self.print_flat_tuple_formula(flat_tuple_formula, print_values_instead_of_labels=True)}" \
                f" = {element_value_to_print}"

    def compute_explain_nested_tuples(self, return_self_if_self_has_mod_obj_container_or_no_ancestors=False):
        if (return_self_if_self_has_mod_obj_container_or_no_ancestors and
                (self.modeling_obj_container is not None or len(self.direct_ancestors_with_id) == 0)):
            return self

        left_explanation = None
        right_explanation = None

        if self.left_parent:
            left_explanation = self.left_parent.compute_explain_nested_tuples(
                return_self_if_self_has_mod_obj_container_or_no_ancestors=True)
        if self.right_parent:
            right_explanation = self.right_parent.compute_explain_nested_tuples(
                return_self_if_self_has_mod_obj_container_or_no_ancestors=True)

        if left_explanation is None and right_explanation is None:
            raise ValueError(f"{self} should have at least one parent to be explained")

        return left_explanation, self.operator, right_explanation

    def compute_formula_as_flat_tuple(self, tuple_element: object) -> tuple:
        if isinstance(tuple_element, ExplainableObject):
            return (tuple_element,)
        elif isinstance(tuple_element, str):
            return (tuple_element,)
        elif isinstance(tuple_element, tuple):
            a, op, b = tuple_element

            if op is None:
                return self.compute_formula_as_flat_tuple(a)

            # Handle unary ops like "X of (Y)"
            if b is None:
                if op is None or len(op) == 0:
                    return self.compute_formula_as_flat_tuple(a)
                return op, '(', *self.compute_formula_as_flat_tuple(a), ')'

            # Determine if we need parentheses
            left_parens = False
            right_parens = False

            if op == "/":
                if isinstance(b, tuple):
                    right_parens = True
                if isinstance(a, tuple) and a[1] != "*":
                    left_parens = True
            elif op == "*":
                if isinstance(a, tuple) and a[1] != "*":
                    left_parens = True
                if isinstance(b, tuple) and b[1] != "*":
                    right_parens = True
            elif op == "-":
                if isinstance(b, tuple) and b[1] in ["+", "-"]:
                    right_parens = True

            left = self.compute_formula_as_flat_tuple(a)
            right = self.compute_formula_as_flat_tuple(b)

            result = []
            if left_parens:
                result += ('(', *left, ')')
            else:
                result += left

            result += (op,)

            if right_parens:
                result += ('(', *right, ')')
            else:
                result += right

            return tuple(result)
        return ()

    @staticmethod
    def print_flat_tuple_formula(flat_formula_tuple: tuple, print_values_instead_of_labels: bool) -> str:
        result = ""

        for el in flat_formula_tuple:
            is_left_paren = el == "("
            is_right_paren = el == ")"
            is_explainable = isinstance(el, ExplainableObject)
            is_operator = not (is_left_paren or is_right_paren or is_explainable)

            if is_explainable:
                s = str(el) if print_values_instead_of_labels else el.label
            else:
                s = str(el)

            # Append with proper spacing rules
            if is_left_paren:
                result += "("
            elif is_right_paren:
                result += ")"
            elif is_operator:
                result += f" {s} "
            else:
                result += s

        return result.lstrip()

    @staticmethod
    def pretty_print_calculation(calc_str):
        return calc_str.replace(" = ", "\n=\n")

    def calculus_graph_to_file(
            self, filename=None, colors_dict=None, x_multiplier=150, y_multiplier=150, width="1800px", height="900px",
            notebook=False, max_depth=100):
        if colors_dict is None:
            colors_dict = {"user data": "gold", "default": "darkred"}
        calculus_graph = build_calculus_graph(
            self, colors_dict, x_multiplier, y_multiplier, width, height, notebook, max_depth=max_depth)

        if filename is None:
            filename = os.path.join(".", f"{self.label} calculus graph.html")

        calculus_graph.show(filename, notebook=notebook)

        add_unique_id_to_mynetwork(filename)

        if notebook:
            from IPython.display import HTML

            return HTML(filename)

        return None

    def serialize_explain_nested_tuples(self):
        localized_explainable_object_type = ExplainableObject  # Localize for faster isinstance

        def recurse(node):
            if node is None:
                return None
            if isinstance(node, tuple):
                left = recurse(node[0])
                right = recurse(node[2])
                return left, node[1], right

            if not isinstance(node, localized_explainable_object_type):
                raise TypeError(f"{node} should be an ExplainableObject but is of type {type(node)}")

            return node.full_str_tuple_id if node.modeling_obj_container is not None else node.to_json()

        return recurse(self.explain_nested_tuples)

    def to_json(self, save_calculated_attributes=False):
        output_dict = {}

        if isinstance(self._value, str):  # Case of technology in WebApplication
            output_dict["value"] = self.value

        output_dict["label"] = self.label

        if self.source is not None:
            output_dict["source"] = {"name": self.source.name, "link": self.source.link}

        if save_calculated_attributes:
            if self._keys_of_direct_ancestors_with_id_loaded_from_json is not None:
                output_dict["direct_ancestors_with_id"] = self._keys_of_direct_ancestors_with_id_loaded_from_json
                output_dict["direct_children_with_id"] = self._keys_of_direct_children_with_id_loaded_from_json
            else:
                output_dict["direct_ancestors_with_id"] = [
                    ancestor.full_str_tuple_id for ancestor in self.direct_ancestors_with_id]
                output_dict["direct_children_with_id"] = [
                    child.full_str_tuple_id for child in self.direct_children_with_id]

            if self.explain_nested_tuples_from_json is not None:
                output_dict["explain_nested_tuples"] = self.explain_nested_tuples_from_json
            else:
                output_dict["explain_nested_tuples"] = self.serialize_explain_nested_tuples()

        return output_dict

    def __repr__(self):
        return str(self)

    def __str__(self):
        self_as_str = str(self.value)
        if len(self_as_str) > 60:
            self_as_str = self_as_str[:60] + "..."

        return self_as_str

    def __eq__(self, other):
        if isinstance(other, ExplainableObject):
            return self.value == other.value

        return False

    def __hash__(self):
        return hash(self.value)



from abc import abstractmethod
from typing import List

from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.builders.services.service_base_class import Service
from efootprint.core.usage.job import JobBase
from efootprint.utils.tools import get_init_signature_params


class ServiceJob(JobBase):
    # Mark the class as abstract but not its children when they define a default_values class attribute
    @classmethod
    @abstractmethod
    def default_values(cls):
        pass

    @classmethod
    def compatible_services(cls):
        init_sig_params = get_init_signature_params(cls)
        service_annotation = init_sig_params["service"].annotation

        return [service_annotation]

    def __init__(self, name: str, service: Service, data_transferred: ExplainableQuantity,
                 data_stored: ExplainableQuantity, request_duration: ExplainableQuantity,
                 compute_needed: ExplainableQuantity, ram_needed: ExplainableQuantity):
        super().__init__(name, data_transferred, data_stored, request_duration, compute_needed, ram_needed)
        self.service = service
        self.ram_needed.set_label(f"RAM needed on server {self.service.server.name} to process {self.name}")
        self.compute_needed.set_label(f"CPU needed on server {self.service.server.name} to process {self.name}")

    @property
    def server(self) -> ModelingObject:
        return self.service.server

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return [self.server] + super().modeling_objects_whose_attributes_depend_directly_on_me



"""Unit mappings for semantic units migration from v12 to v13.

These mappings define which attributes should use which semantic units:
- occurrence: discrete events that are aggregated by sum
- concurrent: concurrent counts that are aggregated by mean
- byte_ram: RAM allocation that is aggregated by mean (vs byte for data transfer)
"""

# Define timeseries attribute mappings: (class_name, attribute_name) -> new_unit_str
# Defined at base class level - automatically applies to all subclasses
TIMESERIES_UNIT_MIGRATIONS = {
    # UsagePattern attributes (applies to all subclasses like EdgeUsagePattern)
    ("UsagePattern", "hourly_usage_journey_starts"): "occurrence",
    ("UsagePattern", "utc_hourly_usage_journey_starts"): "occurrence",
    ("UsagePattern", "nb_usage_journeys_in_parallel"): "concurrent",

    # EdgeUsagePattern specific attributes
    ("EdgeUsagePattern", "hourly_edge_usage_journey_starts"): "occurrence",
    ("EdgeUsagePattern", "utc_hourly_edge_usage_journey_starts"): "occurrence",
    ("EdgeUsagePattern", "nb_edge_usage_journeys_in_parallel"): "concurrent",

    # JobBase attributes (applies to all Job types: Job, GPUJob, etc.)
    ("JobBase", "hourly_occurrences_per_usage_pattern"): "occurrence",
    ("JobBase", "hourly_avg_occurrences_per_usage_pattern"): "concurrent",
    ("JobBase", "hourly_avg_occurrences_across_usage_patterns"): "concurrent",

    # ServerBase attributes (applies to Server, GPUServer, BoaviztaCloudServer, etc.)
    ("ServerBase", "raw_nb_of_instances"): "concurrent",
    ("ServerBase", "nb_of_instances"): "concurrent",

    # Storage attributes
    ("Storage", "raw_nb_of_instances"): "concurrent",
    ("Storage", "nb_of_instances"): "concurrent",
    ("Storage", "nb_of_active_instances"): "concurrent",

    # RecurrentEdgeWorkload attributes (uses ExplainableRecurrentQuantities)
    ("RecurrentEdgeWorkload", "recurrent_workload"): "concurrent",
}

# Define RAM timeseries attributes that need _ram appended (ExplainableHourlyQuantities or ExplainableRecurrentQuantities with byte units)
# Defined at base class level - automatically applies to all subclasses
RAM_TIMESERIES_ATTRIBUTES_TO_MIGRATE = {
    ("RecurrentEdgeProcess", "recurrent_ram_needed"),
    ("RecurrentEdgeProcess", "unitary_hourly_ram_need_per_usage_pattern"),
    ("ServerBase", "hour_by_hour_ram_need"),  # Applies to Server, GPUServer, BoaviztaCloudServer, etc.
    ("EdgeComputer", "unitary_hourly_ram_need_per_usage_pattern")
}

# Define scalar RAM attribute names that need migration: (class_name, attribute_name)
# Defined at base class level - automatically applies to all subclasses
SCALAR_RAM_ATTRIBUTES_TO_MIGRATE = {
    # ServerBase and subclasses (Server, GPUServer, BoaviztaCloudServer, BoaviztaServerFromConfig)
    ("ServerBase", "ram"),
    ("ServerBase", "base_ram_consumption"),
    ("GPUServer", "ram_per_gpu"),  # Specific to GPUServer

    # EdgeComputer
    ("EdgeComputer", "ram"),
    ("EdgeComputer", "base_ram_consumption"),

    # JobBase (applies to Job, GPUJob, etc.)
    ("JobBase", "ram_needed"),

    # Services
    ("VideoStreaming", "base_ram_consumption"),
    ("VideoStreaming", "ram_buffer_per_user"),
}



from efootprint.abstract_modeling_classes.explainable_object_base_class import (
    ExplainableObject, retrieve_update_function_from_mod_obj_and_attr_name)
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.object_linked_to_modeling_obj import ObjectLinkedToModelingObjBase

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject


class ExplainableObjectDict(ObjectLinkedToModelingObjBase, dict):
    """Dict that can be linked to a ModelingObject. Uses ObjectLinkedToModelingObjBase (not slotted)."""

    def __init__(self, input_dict=None):
        super().__init__()
        if input_dict is not None:
            for key, value in input_dict.items():
                self[key] = value

    def set_modeling_obj_container(self, new_parent_modeling_object: ModelingObject, attr_name: str):
        super().set_modeling_obj_container(new_parent_modeling_object, attr_name)
        for value in self.values():
            value.set_modeling_obj_container(new_parent_modeling_object, attr_name)

    @property
    def all_ancestors_with_id(self):
        all_ancestors_with_id = []

        for value in self.values():
            all_ancestor_ids = [ancestor.id for ancestor in all_ancestors_with_id]
            for ancestor in value.all_ancestors_with_id:
                if ancestor.id not in all_ancestor_ids:
                    all_ancestors_with_id.append(ancestor)

        return all_ancestors_with_id

    @property
    def update_function(self):
        if self.modeling_obj_container is None:
            raise ValueError(
                f"{self} doesn’t have a modeling_obj_container, hence it makes no sense "
                f"to look for its update function")
        update_func = retrieve_update_function_from_mod_obj_and_attr_name(
            self.modeling_obj_container, self.attr_name_in_mod_obj_container)

        return update_func

    def __setitem__(self, key, value: ExplainableObject):
        if not isinstance(value, ExplainableObject) and not isinstance(value, EmptyExplainableObject):
            raise ValueError(
                f"ExplainableObjectDicts only accept ExplainableObjects or EmptyExplainableObject as values, "
                f"received {type(value)}")
        if key in self:
            self[key].set_modeling_obj_container(None, None)  # Remove the old modeling object container
        super().__setitem__(key, value)
        value.set_modeling_obj_container(
                new_modeling_obj_container=self.modeling_obj_container, attr_name=self.attr_name_in_mod_obj_container)

    def to_json(self, save_calculated_attributes=False):
        output_dict = {}

        for key, value in self.items():
            if isinstance(key, ModelingObject):
                output_dict[key.id] = value.to_json(save_calculated_attributes)
            elif isinstance(key, str):
                output_dict[key] = value.to_json(save_calculated_attributes)
            else:
                raise ValueError(f"Key {key} is not a ModelingObject or a string")

        return output_dict

    def __repr__(self):
        return str(self)

    def __str__(self):
        if len(self) == 0:
            return "{}"

        return_str = "{\n"

        for key, value in self.items():
            if isinstance(key, ModelingObject):
                return_str += f"{key.id}: {value}, \n"
            elif isinstance(key, str):
                return_str += f"{key}: {value}, \n"
            else:
                raise ValueError(f"Key {key} is not a ModelingObject or a string")

        return_str = return_str + "}"

        return return_str



from typing import TYPE_CHECKING

from efootprint.abstract_modeling_classes.explainable_recurrent_quantities import ExplainableRecurrentQuantities
from efootprint.core.hardware.edge.edge_storage import EdgeStorage
from efootprint.core.usage.edge.recurrent_edge_component_need import RecurrentEdgeComponentNeed

if TYPE_CHECKING:
    from efootprint.core.usage.edge.edge_usage_pattern import EdgeUsagePattern


class RecurrentEdgeStorageNeed(RecurrentEdgeComponentNeed):
    def __init__(self, name: str, edge_component: EdgeStorage, recurrent_need: ExplainableRecurrentQuantities):
        super().__init__(name, edge_component, recurrent_need)

    def update_dict_element_in_unitary_hourly_need_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        # First compute the base hourly need using parent logic
        super().update_dict_element_in_unitary_hourly_need_per_usage_pattern(usage_pattern)

        # Get the computed value
        base_storage_need = self.unitary_hourly_need_per_usage_pattern[usage_pattern]

        # Apply Monday 00:00 logic
        # if usage_pattern.nb_edge_usage_journey_in_parallel.start_date doesn't start on a Monday 00:00,
        # set the first values of the storage need to 0 until the first Monday 00:00, so that if storage need increases
        # during beginning of the week then decreases at the end of the week, it doesn't go negative
        start_date_weekday = usage_pattern.nb_edge_usage_journeys_in_parallel.start_date.weekday()
        start_date_hour = usage_pattern.nb_edge_usage_journeys_in_parallel.start_date.hour
        if start_date_weekday != 0 or start_date_hour != 0:
            hours_until_first_monday_00 = (7 - start_date_weekday) * 24 - start_date_hour
            base_storage_need.magnitude[:hours_until_first_monday_00] = 0

        # Re-set with updated label
        self.unitary_hourly_need_per_usage_pattern[usage_pattern] = base_storage_need.set_label(
            f"{self.name} unitary hourly need for {usage_pattern.name}")



from copy import copy
from datetime import datetime, timedelta
from time import perf_counter
from typing import List

from efootprint.abstract_modeling_classes.contextual_modeling_object_attribute import ContextualModelingObjectAttribute
from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, \
    optimize_attr_updates_chain
from efootprint.abstract_modeling_classes.object_linked_to_modeling_obj import (
    ObjectLinkedToModelingObj, ObjectLinkedToModelingObjBase)
from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject, optimize_mod_objs_computation_chain
from efootprint.logger import logger


def compute_attr_updates_chain_from_mod_objs_computation_chain(mod_objs_computation_chain: List[ModelingObject]):
    attr_updates_chain = []
    for mod_obj in mod_objs_computation_chain:
        for calculated_attribute in mod_obj.calculated_attributes:
            attr_updates_chain.append(getattr(mod_obj, calculated_attribute))

    return attr_updates_chain


class ModelingUpdate:
    def __init__(
            self, changes_list: List[List[ObjectLinkedToModelingObj | list | dict]], simulation_date: datetime = None,
            compute_previous_system_footprints=True):
        start = perf_counter()
        self.system = None
        for change in changes_list:
            changed_val = change[0]
            if isinstance(changed_val, ObjectLinkedToModelingObjBase) and changed_val.modeling_obj_container.systems:
                self.system = changed_val.modeling_obj_container.systems[0]
                break
        self.changes_list = changes_list
        self.parse_changes_list()
        if self.changes_list and self.system and compute_previous_system_footprints:
            self.system.previous_total_energy_footprints_sum_over_period = (
                self.system.total_energy_footprint_sum_over_period)
            self.system.previous_total_fabrication_footprints_sum_over_period = \
                self.system.total_fabrication_footprint_sum_over_period
            self.system.previous_change = changes_list
            self.system.all_changes += changes_list

        self.simulation_date = simulation_date
        if simulation_date is not None:
            assert self.system is not None
            if simulation_date.tzinfo is None:
                raise ValueError(
                    f"Simulation date {simulation_date} should be timezone aware. "
                    f"Please use a timezone aware datetime object by setting its tzinfo attribute.")
            self.system.simulation = self

        self.mod_objs_computation_chain = self.compute_mod_objs_computation_chain()
        if self.mod_objs_computation_chain:
            logger.info(f"{len(self.mod_objs_computation_chain)} recomputed objects: "
                        f"{[mod_obj.name for mod_obj in self.mod_objs_computation_chain]}")
        self.attr_updates_chain_from_mod_objs_computation_chains = (
            compute_attr_updates_chain_from_mod_objs_computation_chain(self.mod_objs_computation_chain))
        self.values_to_recompute = self.generate_optimized_attr_updates_chain()

        self.ancestors_not_in_computation_chain = []
        self.hourly_quantities_to_filter = []
        self.filtered_hourly_quantities = []
        self.ancestors_to_replace_by_copies = []
        self.replaced_ancestors_copies = []
        if self.simulation_date is not None:
            self.make_simulation_specific_operations()
            logger.info("Simulation specific operations done.")

        self.apply_changes()
        for new_sourcevalue in self.new_sourcevalues:
            mod_obj_container = new_sourcevalue.modeling_obj_container
            mod_obj_container.check_belonging_to_authorized_values(
                new_sourcevalue.attr_name_in_mod_obj_container, new_sourcevalue,
                mod_obj_container.attributes_with_depending_values())
        self.recomputed_values = []
        self.updated_values_set = True
        try:
            self.recompute_attributes()
        except Exception as e:
            logger.error("An error occurred during attribute recomputation. Resetting to previous values.")
            self.reset_values()
            e.args = (f"Error occurred while computing changes. All changes have been reset."
                      f"\nOriginal error:\n {e}",) + e.args[1:]
            raise e

        if self.simulation_date is not None:
            self.link_simulated_and_baseline_twins()

        if simulation_date is not None:
            self.reset_values()
        compute_time_ms = round(1000 * (perf_counter() - start), 1)
        avg_compute_time_per_value = round(compute_time_ms / len(self.values_to_recompute), 2)\
            if self.values_to_recompute else 0
        logger.info(f"{len(self.changes_list)} changes lead to {len(self.values_to_recompute)} update computations "
                    f"done in {compute_time_ms} ms (avg {avg_compute_time_per_value} ms per computation).")

    @property
    def previous_and_new_objects_organized_in_sections(self):
        return [
            ["direct changes", [change[0] for change in self.changes_list], [change[1] for change in self.changes_list]],
            ["filtered hourly quantities", self.hourly_quantities_to_filter, self.filtered_hourly_quantities],
            ["replaced ancestors copies", self.ancestors_to_replace_by_copies, self.replaced_ancestors_copies],
            ["recomputed values", self.values_to_recompute, self.recomputed_values]
        ]

    def parse_changes_list(self):
        indexes_to_skip = []
        index = 0
        while index < len(self.changes_list):
            old_value, new_value = self.changes_list[index]
            assert isinstance(old_value, ObjectLinkedToModelingObjBase), \
                              f"{old_value} should be an ObjectLinkedToModelingObjBase but is of type {type(old_value)}"
            if new_value is None:
                assert isinstance(old_value, ExplainableObject)
                self.changes_list[index][1] = EmptyExplainableObject()
            else:
                mod_obj_container = old_value.modeling_obj_container
                mod_obj_container.check_input_value_type_positivity_and_unit(
                    old_value.attr_name_in_mod_obj_container, new_value)

            if isinstance(new_value, list):
                from efootprint.abstract_modeling_classes.list_linked_to_modeling_obj import ListLinkedToModelingObj
                self.changes_list[index][1] = ListLinkedToModelingObj(new_value)
            if isinstance(new_value, ModelingObject):
                self.changes_list[index][1] = ContextualModelingObjectAttribute(new_value)
                if old_value.attr_name_in_mod_obj_container in mod_obj_container.attribute_update_entanglements:
                    changes_to_add = mod_obj_container.attribute_update_entanglements[
                        old_value.attr_name_in_mod_obj_container]([old_value, new_value])
                    self.changes_list += changes_to_add

            if not isinstance(self.changes_list[index][1], ObjectLinkedToModelingObjBase):
                raise ValueError(
                    f"New e-footprint attributes should be ObjectLinkedToModelingObjBase,"
                    f" got {old_value} of type {type(old_value)} trying to be set to an object "
                    f"of type {type(new_value)}")

            if old_value == new_value:
                logger.warning(
                    f"{old_value.id} is updated to itself. "
                    f"It happens when using my_mod_obj.list_attribute += other list syntax. "
                    f"Otherwise This is surprising, you might want to double check your action. "
                    f"The link update logic will be skipped.")
                indexes_to_skip.append(index)
            index += 1

        for index in sorted(indexes_to_skip, reverse=True):
            del self.changes_list[index]

    def compute_mod_objs_computation_chain(self):
        from efootprint.abstract_modeling_classes.list_linked_to_modeling_obj import ListLinkedToModelingObj
        mod_objs_computation_chain = []
        for old_value, new_value in self.changes_list:
            if isinstance(old_value, ContextualModelingObjectAttribute):
                mod_objs_computation_chain += (
                    old_value.modeling_obj_container.compute_mod_objs_computation_chain_from_old_and_new_modeling_objs(
                        old_value, new_value, optimize_chain=False))
            elif isinstance(old_value, ListLinkedToModelingObj):
                mod_objs_computation_chain += (
                    old_value.modeling_obj_container.compute_mod_objs_computation_chain_from_old_and_new_lists(
                        old_value, new_value, optimize_chain=False))

        optimized_chain = optimize_mod_objs_computation_chain(mod_objs_computation_chain)

        return optimized_chain

    def apply_changes(self):
        for old_value, new_value in self.changes_list:
            old_value.replace_in_mod_obj_container_without_recomputation(new_value)

    def make_simulation_specific_operations(self):
        assert self.simulation_date is not None
        self.ancestors_not_in_computation_chain = self.compute_ancestors_not_in_computation_chain()
        self.hourly_quantities_to_filter = self.compute_hourly_quantities_to_filter()
        if self.mod_objs_computation_chain:
            # The simulation will change the calculation graph, so we need to replace all ancestors not in
            # computation chain by their copies to keep the original calculation graph unchanged
            self.ancestors_to_replace_by_copies = [
                ancestor for ancestor in self.ancestors_not_in_computation_chain
                if ancestor.id not in [value.id for value in self.hourly_quantities_to_filter]]
            self.replaced_ancestors_copies = self.replace_ancestors_not_in_computation_chain_by_copies()
        self.filter_hourly_quantities_to_filter()

    def recompute_attributes(self):
        for value_to_recompute in self.values_to_recompute:
            attr_name_in_mod_obj_container = value_to_recompute.attr_name_in_mod_obj_container
            modeling_obj_container = value_to_recompute.modeling_obj_container
            key_in_dict = None
            if value_to_recompute.dict_container is not None:
                key_in_dict = value_to_recompute.key_in_dict
            if not key_in_dict:
                logger.debug(f"Recomputing {attr_name_in_mod_obj_container} in {modeling_obj_container.id}")
                value_to_recompute.update_function()
                recomputed_value = getattr(modeling_obj_container, attr_name_in_mod_obj_container)
            else:
                logger.debug(f"Recomputing {attr_name_in_mod_obj_container} in {modeling_obj_container.id} "
                             f"with key {key_in_dict.id}")
                value_to_recompute.update_function(key_in_dict)
                recomputed_value = getattr(modeling_obj_container, attr_name_in_mod_obj_container)[key_in_dict]
            self.recomputed_values.append(recomputed_value)

    @property
    def old_sourcevalues(self):
        return [old_value for old_value, new_value in self.changes_list if isinstance(old_value, ExplainableObject)]

    @property
    def new_sourcevalues(self):
        return [new_value for old_value, new_value in self.changes_list if isinstance(old_value, ExplainableObject)]

    def generate_optimized_attr_updates_chain(self):
        attr_updates_chain_from_attributes_updates = sum(
            [old_value.attr_updates_chain for old_value in self.old_sourcevalues], start=[])

        # Necessary to do the sum in this order because calculations from modeling objects computation chains must be
        # done after the calculations from input updates.
        optimized_chain = optimize_attr_updates_chain(
            attr_updates_chain_from_attributes_updates + self.attr_updates_chain_from_mod_objs_computation_chains)

        optimized_chain_without_previous_nor_initial_values = [
            attr for attr in optimized_chain if not attr.attr_name_in_mod_obj_container.startswith("previous_")
                                                and not attr.attr_name_in_mod_obj_container.startswith("initial_")]

        return optimized_chain_without_previous_nor_initial_values

    def compute_ancestors_not_in_computation_chain(self):
        all_ancestors_of_values_to_recompute = sum(
            [value.all_ancestors_with_id for value in self.values_to_recompute], start=[])
        deduplicated_all_ancestors_of_values_to_recompute = []
        for ancestor in all_ancestors_of_values_to_recompute:
            if ancestor.id not in [elt.id for elt in deduplicated_all_ancestors_of_values_to_recompute]:
                deduplicated_all_ancestors_of_values_to_recompute.append(ancestor)
        values_to_recompute_attribute_ids = [elt.attribute_id for elt in self.values_to_recompute]
        old_sourcevalues_attribute_ids = [old_value.attribute_id for old_value in self.old_sourcevalues]
        ancestors_not_in_computation_chain = [
            ancestor for ancestor in deduplicated_all_ancestors_of_values_to_recompute
            if ancestor.attribute_id not in values_to_recompute_attribute_ids + old_sourcevalues_attribute_ids]

        return ancestors_not_in_computation_chain

    def compute_hourly_quantities_to_filter(self):
        hourly_quantities_ancestors_not_in_computation_chain = [
            ancestor for ancestor in self.ancestors_not_in_computation_chain
            if isinstance(ancestor, ExplainableHourlyQuantities)
        ]

        hourly_quantities_to_filter = []
        global_min_date = None
        global_max_date = None

        for ancestor in hourly_quantities_ancestors_not_in_computation_chain:
            start = ancestor.start_date
            end = start + timedelta(hours=len(ancestor.value) - 1)

            # Ensure timezone awareness
            if start.tzinfo is None:
                # Should only be the case for UsagePattern’s hourly_usage_journeys
                start = start.replace(tzinfo=ancestor.modeling_obj_container.country.timezone.value)
            if end.tzinfo is None:
                end = end.replace(tzinfo=ancestor.modeling_obj_container.country.timezone.value)

            # Track global range
            if global_min_date is None:
                global_min_date = start
            else:
                global_min_date = min(global_min_date, start)

            if global_max_date is None:
                global_max_date = end
            else:
                global_max_date = max(global_max_date, end)

            # Filtering condition
            if self.simulation_date <= end:
                hourly_quantities_to_filter.append(ancestor)

        # Final consistency check
        if not (global_min_date <= self.simulation_date <= global_max_date):
            raise ValueError(
                f"Can't start a simulation on {self.simulation_date} because "
                f"{self.simulation_date} doesn't belong to the existing modeling period "
                f"from {global_min_date} to {global_max_date}"
            )

        return hourly_quantities_to_filter

    def filter_hourly_quantities_to_filter(self):
        for hourly_quantities in self.hourly_quantities_to_filter:
            start = hourly_quantities.start_date
            if start.tzinfo is None:
                start = hourly_quantities.modeling_obj_container.country.timezone.value.localize(start)

            # Find positions at or after simulation_date
            mask = [start + timedelta(hours=i) >= self.simulation_date for i in range(len(hourly_quantities.value))]
            filtered_values = hourly_quantities.value[mask]

            if len(filtered_values) == 0:
                new_value = EmptyExplainableObject()
            else:
                new_value = ExplainableHourlyQuantities(
                    filtered_values,
                    start_date=self.simulation_date,
                    label=hourly_quantities.label,
                    left_parent=hourly_quantities.left_parent,
                    right_parent=hourly_quantities.right_parent,
                    operator=hourly_quantities.operator,
                    source=hourly_quantities.source,
                )

            hourly_quantities.replace_in_mod_obj_container_without_recomputation(new_value)
            self.filtered_hourly_quantities.append(new_value)

    def replace_ancestors_not_in_computation_chain_by_copies(self):
        copies = []
        for ancestor_to_replace_by_copy in self.ancestors_to_replace_by_copies:
            # Replace all ancestors not in computation chain by their copy so that the original calculation graph
            # will remain unchanged when the simulation is over
            ancestor_copy = copy(ancestor_to_replace_by_copy)
            ancestor_copy.left_parent = None
            ancestor_copy.right_parent = None
            ancestor_copy.operator = None
            ancestor_to_replace_by_copy.replace_in_mod_obj_container_without_recomputation(ancestor_copy)
            copies.append(ancestor_copy)

        return copies

    def reset_values(self):
        if self.updated_values_set:
            for section_name, previous_values, new_values in self.previous_and_new_objects_organized_in_sections:
                logger.info(f"Resetting {section_name} from {len(new_values)} updated values")
                for new_value, previous_value in zip(new_values, previous_values):
                    new_value.replace_in_mod_obj_container_without_recomputation(previous_value)
                self.updated_values_set = False

    def set_updated_values(self):
        if not self.updated_values_set:
            for section_name, previous_values, new_values in self.previous_and_new_objects_organized_in_sections:
                logger.info(f"Setting {section_name} from {len(previous_values)} previous values")
                for new_value, previous_value in zip(new_values, previous_values):
                    previous_value.replace_in_mod_obj_container_without_recomputation(new_value)
                self.updated_values_set = True

    def link_simulated_and_baseline_twins(self):
        assert self.simulation_date is not None
        for value_to_recompute, recomputed_value in zip(self.values_to_recompute, self.recomputed_values):
            value_to_recompute.simulation_twin = recomputed_value
            recomputed_value.baseline_twin = value_to_recompute
            value_to_recompute.simulation = self
            recomputed_value.simulation = self



from copy import copy

import numpy as np
from pint import Quantity

from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
from efootprint.utils.plot_baseline_and_simulation_data import plot_baseline_and_simulation_data, prepare_data


@ExplainableObject.register_subclass(lambda d: "value" in d and d["value"] is None)
class EmptyExplainableObject(ExplainableObject):
    __slots__ = (
        '_ExplainableQuantity',
        '_ExplainableHourlyQuantities',
    )

    @classmethod
    def from_json_dict(cls, d):
        return cls(label=d["label"])

    def __init__(self, label="no value", left_parent: ExplainableObject = None, right_parent: ExplainableObject = None,
                 operator: str = None):
        from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
        from efootprint.abstract_modeling_classes.explainable_hourly_quantities import ExplainableHourlyQuantities
        self._ExplainableQuantity = ExplainableQuantity
        self._ExplainableHourlyQuantities = ExplainableHourlyQuantities
        super().__init__(
            value=None, label=label, left_parent=left_parent, right_parent=right_parent, operator=operator)
        self.value = self

    def to(self, unit):
        return self

    def check(self, str_unit):
        return True

    def ceil(self):
        return EmptyExplainableObject(left_parent=self, operator="ceil")

    def max(self):
        return EmptyExplainableObject(left_parent=self, operator="max")

    def abs(self):
        return EmptyExplainableObject(left_parent=self, operator="abs")

    def sum(self):
        return EmptyExplainableObject(left_parent=self, operator="sum")

    def copy(self):
        return EmptyExplainableObject(left_parent=self, operator="copy")

    def generate_explainable_object_with_logical_dependency(self, explainable_condition: ExplainableObject):
        return EmptyExplainableObject(
            label=self.label, left_parent=self, right_parent=explainable_condition, operator="logically dependent on")

    @property
    def magnitude(self):
        return 0

    def __copy__(self):
        return EmptyExplainableObject(label=copy(self.label))

    def __eq__(self, other):
        if isinstance(other, EmptyExplainableObject):
            return True
        elif isinstance(other, ExplainableObject):
            return other.__eq__(self)
        elif other == 0:
            return True

        return False

    def __round__(self, round_level):
        return EmptyExplainableObject(
            label=self.label, left_parent=self, operator=f"rounded to {round_level} decimals")

    def __add__(self, other):
        if isinstance(other, EmptyExplainableObject):
            return EmptyExplainableObject(left_parent=self, right_parent=other, operator="+")
        if isinstance(other, ExplainableObject):
            return other.__add__(self)
        elif other == 0:
            return EmptyExplainableObject(left_parent=self, operator="+ 0")
        else:
            raise ValueError

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, EmptyExplainableObject):
            return EmptyExplainableObject(left_parent=self, right_parent=other, operator="-")
        else:
            raise ValueError

    def __mul__(self, other):
        if isinstance(other, EmptyExplainableObject):
            return EmptyExplainableObject(left_parent=self, right_parent=other, operator="*")
        if isinstance(other, self._ExplainableQuantity) or isinstance(other, self._ExplainableHourlyQuantities):
            return other.__mul__(self)
        elif other == 0:
            return self
        else:
            raise ValueError

    def __rmul__(self, other):
        return self.__mul__(other)

    def __str__(self):
        return "no value"

    def __deepcopy__(self, memo):
        return EmptyExplainableObject(label=self.label, left_parent=self.left_parent, right_parent=self.right_parent,
                                      operator=self.operator)

    def np_compared_with(self, compared_object, comparator):
        if isinstance(compared_object, EmptyExplainableObject):
            return EmptyExplainableObject(left_parent=self, right_parent=compared_object,
                                          operator=f"{comparator} compared with")
        elif isinstance(compared_object, self._ExplainableHourlyQuantities):
            return compared_object.np_compared_with(self, comparator)
        else:
            raise ValueError(f"Can only compare with another EmptyExplainableObject or ExplainableHourlyQuantities,"
                             f" not {type(compared_object)}")

    def to_json(self, save_calculated_attributes=False):
        output_dict = {"value": None}
        output_dict.update(super().to_json(save_calculated_attributes))

        return output_dict

    def plot(self, figsize=(10, 4), filepath=None, plt_show=False, xlims=None, cumsum=False):
        import matplotlib.pyplot as plt

        assert self.simulation_twin is not None, \
            "Cannot plot EmptyExplainableObject if simulation twin is None"

        sim_obj = self.simulation_twin
        assert not isinstance(sim_obj, EmptyExplainableObject), \
            "Cannot plot EmptyExplainableObject if simulation twin is EmptyExplainableObject"

        time_sim, simulated_q = prepare_data(sim_obj.value, sim_obj.start_date, apply_cumsum=cumsum)
        unit = simulated_q.units

        # Create a baseline of zeros with the same time index and unit
        baseline_q = Quantity(np.zeros_like(simulated_q.magnitude, dtype=np.float32), unit)
        time_baseline = time_sim

        fig, ax = plot_baseline_and_simulation_data(
            baseline_q, time_baseline, simulated_q, time_sim, figsize=figsize, xlims=xlims)

        if self.label:
            title = self.label if not cumsum else f"Cumulative {self.label[0].lower()}{self.label[1:]}"
            ax.set_title(title)

        if filepath is not None:
            plt.savefig(filepath, bbox_inches='tight')

        if plt_show:
            plt.show()

        return fig, ax
